from spaceone.core.auth import Authenticator
from spaceone.core.auth.jwt.jwt_util import JWTUtil
from spaceone.core.error import *


class JWTAuthenticator(Authenticator):
    def __init__(self, key):
        self._key = key

    def validate(self, token, options=None):
        if not self._key:
            raise ERROR_AUTHENTICATE_FAILURE(message="Decode key is not set.")

        if not isinstance(token, (str, bytes)):
            raise ERROR_AUTHENTICATE_FAILURE(message="Invalid token format.")

        if options is None:
            options = {}

        try:
            payload = JWTUtil.decode(token, self._key, options=options)
        except Exception:
            raise ERROR_AUTHENTICATE_FAILURE(message="Token is invalid or expired.")

        return payload



from spaceone.core import config
from spaceone.core.logger import set_logger
from spaceone.core.opentelemetry import set_tracer, set_metric
from spaceone.core.pygrpc.server import GRPCServer, add_extension_services


class PluginServer(object):

    _grpc_app: GRPCServer = None
    _global_conf_path: str = None
    _plugin_methods = {}

    def __init__(self):
        if self._global_conf_path:
            config.set_service_config(self._global_conf_path)

        # Enable logging configuration
        set_logger()

        # Set OTel Tracer and Metric
        set_tracer()

    @property
    def grpc_app(self) -> GRPCServer:
        return self._grpc_app

    @classmethod
    def route(cls, plugin_method: str) -> callable:
        def wrapper(func: callable):
            class_name, method_name = plugin_method.split('.')

            if class_name not in cls._plugin_methods:
                raise ValueError(f'Plugin method is invalid. (plugin_method = {plugin_method}, '
                                 f'allowed = {list(cls._plugin_methods.keys())})')

            if method_name not in cls._plugin_methods[class_name]['methods']:
                raise ValueError(f'Plugin method is invalid. (plugin_method = {plugin_method},'
                                 f'allowed = {list(cls._plugin_methods[class_name]["methods"])})')

            cls._plugin_methods[class_name]['service'].set_plugin_method(method_name, func)

        return wrapper

    def run(self):
        add_extension_services(self.grpc_app)
        self.grpc_app.run()


def _get_plugin_app() -> PluginServer:
    package: str = config.get_package()
    app_path: str = config.get_global('PLUGIN_APP_PATH')
    app_path = app_path.format(package=package)
    module_path, app_name = app_path.split(':')

    try:
        app_module = __import__(module_path, fromlist=[app_name])

        if not hasattr(app_module, 'app'):
            raise ImportError(f'App is not defined. (app_path = {package}.{app_path})')

        return getattr(app_module, 'app')
    except Exception as e:
        raise ImportError(f'Cannot import app: {e}')


def serve(app_path: str = None):
    app = _get_plugin_app()
    app.run()


import copy
import functools
import logging
import time
from typing import Generator, Union, Literal

from opentelemetry import trace
from opentelemetry.trace import format_trace_id
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from spaceone.core.base import CoreObject
from spaceone.core import config
from spaceone.core.error import *
from spaceone.core.service.utils import *
from spaceone.core.transaction import (
    get_transaction,
    create_transaction,
    delete_transaction,
)
from spaceone.core.handler import *

_HANDLER_TYPE = Literal["authentication", "authorization", "mutation", "event"]
_ERROR_TYPE = Literal["INVALID_ARGUMENT", "UNKNOWN"]
_LOGGER = logging.getLogger(__name__)
_TRACER = trace.get_tracer(__name__)

__all__ = [
    "BaseService",
    "transaction",
    "authorization_handler",
    "authentication_handler",
    "mutation_handler",
    "event_handler",
    "convert_model",
    "change_value_by_rule",
    "change_only_key",
    "check_required",
    "set_query_page_limit",
    "append_query_filter",
    "change_tag_filter",
    "append_keyword_filter",
    "change_timestamp_value",
    "change_date_value",
    "change_timestamp_filter",
]
_METADATA_KEYS = [
    "token",
    "x_domain_id",
    "x_workspace_id",
    "traceparent",
    "peer",
]


class BaseService(CoreObject):
    service = None
    resource = None
    _plugin_methods = {}
    _handler_state = {}

    def __init__(self, metadata: dict = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_span_context = None
        self.service = config.get_service()
        self._metadata = {}
        self._set_metadata(metadata)

    def _set_metadata(self, metadata: dict = None) -> None:
        metadata = metadata or self.transaction.meta
        for key in _METADATA_KEYS:
            if value := metadata.get(key):
                self._metadata[key] = value

    @property
    def metadata(self) -> dict:
        return self._metadata

    @classmethod
    def set_plugin_method(cls, method_name: str, func: callable) -> None:
        cls._plugin_methods[method_name] = func

    @classmethod
    def get_plugin_method(cls, method_name: str) -> Union[callable, None]:
        return cls._plugin_methods.get(method_name)

    @classmethod
    def enable_handler(cls, handler_type: _HANDLER_TYPE) -> None:
        if cls._handler_state.get(cls.__name__) is None:
            cls._handler_state[cls.__name__] = {
                "authentication": False,
                "authorization": False,
                "mutation": False,
                "event": False,
            }

        cls._handler_state[cls.__name__][handler_type] = True

    @classmethod
    def get_handler_state(cls) -> dict:
        return cls._handler_state.get(
            cls.__name__,
            {
                "authentication": False,
                "authorization": False,
                "mutation": False,
                "event": False,
            },
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def transaction(
    cls_func: callable = None,
    permission: str = None,
    role_types: list = None,
    exclude: list = None,
    verb: str = None,
) -> callable:
    def wrapper(func: callable):
        @functools.wraps(func)
        def wrapped_func(self, params: dict):
            _verb = verb or func.__name__
            _traceparent = self.metadata.get("traceparent")

            with _TRACER.start_as_current_span(
                f"{self.resource}.{_verb}", context=_get_span_context(_traceparent)
            ) as span:
                self.current_span_context = span.get_span_context()
                trace_id = format_trace_id(self.current_span_context.trace_id)
                create_transaction(
                    self.service, self.resource, _verb, trace_id, self.metadata
                )
                return _pipeline(
                    func,
                    self,
                    params,
                    permission,
                    role_types,
                    exclude or [],
                    self.get_handler_state(),
                )

        return wrapped_func

    return wrapper(cls_func) if cls_func else wrapper


def _get_span_context(traceparent: str = None):
    if traceparent:
        carrier = {"traceparent": traceparent}
        return TraceContextTextMapPropagator().extract(carrier)
    else:
        return None


def _pipeline(
    func: callable,
    self: BaseService,
    params: Union[Generator, dict, None],
    permission: str,
    role_types: list,
    exclude: list,
    handler_state: dict,
) -> Union[Generator, dict, None]:
    print_info_log = (
        str(self.metadata.get("disable_info_log", "false")).lower() != "true"
    )

    authentication_handler_state = (
        handler_state["authentication"] and "authentication" not in exclude
    )
    authorization_handler_state = (
        handler_state["authorization"] and "authorization" not in exclude
    )
    mutation_handler_state = handler_state["mutation"] and "mutation" not in exclude
    event_handler_state = handler_state["event"] and "event" not in exclude

    try:
        with _TRACER.start_as_current_span("PreProcessing"):
            start_time = time.time()

            # 1. Event - Start
            if event_handler_state:
                for handler in get_event_handlers():
                    with _TRACER.start_as_current_span(handler.__class__.__name__):
                        handler.notify("STARTED", params)

            # 2. Authentication
            if authentication_handler_state:
                for handler in get_authentication_handlers():
                    with _TRACER.start_as_current_span(handler.__class__.__name__):
                        handler.verify(params)

            # 3. Authorization
            if authorization_handler_state:
                for handler in get_authorization_handlers():
                    with _TRACER.start_as_current_span(handler.__class__.__name__):
                        handler.verify(params, permission, role_types)

            # 4. Print Request Info Log
            if print_info_log:
                _LOGGER.info("(REQUEST) =>", extra={"parameter": copy.deepcopy(params)})

            # 5. Request Mutation
            if mutation_handler_state:
                for handler in get_mutation_handlers():
                    with _TRACER.start_as_current_span(handler.__class__.__name__):
                        params = handler.request(params)

        # 6. Service Body
        with _TRACER.start_as_current_span(
            f"Body", links=[trace.Link(self.current_span_context)]
        ):
            # 6. Event - In Progress
            if event_handler_state:
                for handler in get_event_handlers():
                    with _TRACER.start_as_current_span(handler.__class__.__name__):
                        handler.notify("IN_PROGRESS", params)

            response_or_iterator = func(self, params)
            if isinstance(response_or_iterator, Generator):
                # Skip PostProcessing
                return response_or_iterator

        with _TRACER.start_as_current_span("PostProcessing"):
            # 7. Response Mutation
            if mutation_handler_state:
                for handler in get_mutation_handlers(reverse=True):
                    with _TRACER.start_as_current_span(handler.__class__.__name__):
                        response_or_iterator = handler.response(response_or_iterator)

            # 8. Event - Success
            if event_handler_state:
                for handler in get_event_handlers():
                    with _TRACER.start_as_current_span(handler.__class__.__name__):
                        handler.notify("SUCCESS", response_or_iterator)

            # 9. Print Response Info Log
            if print_info_log:

                process_time = time.time() - start_time
                _LOGGER.info(f"(RESPONSE) => SUCCESS (Time = {process_time:.2f}s)")

        return response_or_iterator

    except ERROR_INVALID_ARGUMENT as e:
        _error_handler("INVALID_ARGUMENT", e, event_handler_state)
        raise e

    except ERROR_BASE as e:
        _error_handler("UNKNOWN", e, event_handler_state)
        raise e

    except Exception as e:
        error = ERROR_UNKNOWN(message=e)
        _error_handler("UNKNOWN", error, event_handler_state)
        raise error

    finally:
        delete_transaction()


def _error_handler(
    error_type: _ERROR_TYPE,
    error: ERROR_BASE,
    event_handler_state: bool,
) -> None:
    error.meta["skip_error_log"] = True

    # Event - Failure
    if event_handler_state:
        for handler in get_event_handlers():
            handler.notify(
                "FAILURE", {"error_code": error.error_code, "message": error.message}
            )

    _LOGGER.error(f"(Error) => {error.message} {error}", exc_info=True)

    if tnx := get_transaction(is_create=False):
        tnx.execute_rollback()


def authentication_handler(cls: BaseService):
    return _set_handler(cls, "authentication")


def authorization_handler(cls: BaseService):
    return _set_handler(cls, "authorization")


def mutation_handler(cls: BaseService):
    return _set_handler(cls, "mutation")


def event_handler(cls: BaseService):
    return _set_handler(cls, "event")


def _set_handler(cls: BaseService, handler_type: _HANDLER_TYPE):
    cls.enable_handler(handler_type)
    return cls



from xmlrunner import XMLTestRunner
from spaceone.core.unittest.result import RichTestResult

UTF8 = 'UTF-8'


class RichTestRunner(XMLTestRunner):
    """
    A test runner class that outputs the results in JUnit like XML files.
    """

    def __init__(self, output='reports', outsuffix=None,
                 elapsed_times=True, encoding=UTF8,
                 resultclass=RichTestResult, 
                 **kwargs):
        super(RichTestRunner, self).__init__(output, outsuffix,
                                             elapsed_times, encoding, resultclass, **kwargs)



import json
from jwcrypto import jwk
from jose import jwt


class JWTUtil:
    @staticmethod
    def generate_jwk(key_type="RSA", size=2048):
        key = jwk.JWK.generate(kty=key_type, size=size)
        private_jwk = json.loads(key.export_private())
        public_jwk = json.loads(key.export_public())
        return private_jwk, public_jwk

    @staticmethod
    def encode(payload: dict, private_jwk: dict, algorithm="RS256") -> str:
        return jwt.encode(payload, key=private_jwk, algorithm=algorithm)

    @staticmethod
    def decode(token: str, public_jwk: dict, algorithm="RS256", options=None) -> dict:
        if options is None:
            options = {}

        options["verify_aud"] = options.get("verify_aud", False)

        return jwt.decode(token, key=public_jwk, algorithms=algorithm, options=options)

    @staticmethod
    def unverified_decode(token: str) -> dict:
        return jwt.get_unverified_claims(token)

    @staticmethod
    def get_value_from_token(token: str, key: str, default: any = None) -> any:
        try:
            return JWTUtil.unverified_decode(token).get(key, default)
        except Exception as e:
            return default



import types
import logging
from typing import Any, List, Tuple
from google.protobuf.json_format import MessageToDict
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from spaceone.core.connector import BaseConnector
from spaceone.core import pygrpc
from spaceone.core.utils import parse_grpc_endpoint
from spaceone.core.pygrpc.client import GRPCClient
from spaceone.core.error import *

__all__ = ["SpaceConnector"]

_LOGGER = logging.getLogger(__name__)


class SpaceConnector(BaseConnector):
    name = "SpaceConnector"

    def __init__(
        self,
        *args,
        service: str = None,
        endpoint: str = None,
        token: str = None,
        return_type: str = "dict",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._service = service
        self._endpoint = endpoint
        self._token = token
        self._return_type = return_type

        self._client = None
        self._endpoints: dict = self.config.get("endpoints", {})

        self._verify()
        self._init_client()

    @property
    def client(self) -> GRPCClient:
        return self._client

    def dispatch(self, method: str, params: dict = None, **kwargs) -> Any:
        _LOGGER.debug(f"space_connector.dispatch : method: {method}")
        return self._call_api(method, params, **kwargs)

    def _call_api(
        self,
        method: str,
        params: dict = None,
        token: str = None,
        x_domain_id: str = None,
        x_workspace_id: str = None,
    ) -> Any:
        resource, verb = self._parse_method(method)
        _LOGGER.debug(f"space_connector.dispatch :: resource: {resource}, verb: {verb}")

        self._check_method(resource, verb)

        params = params or {}
        _LOGGER.debug(f"space_connector.dispatch :: params: {params}")
        metadata = self._get_connection_metadata(token, x_domain_id, x_workspace_id)
        _LOGGER.debug(f"space_connector.dispatch :: metadata: {metadata}")

        response_or_iterator = getattr(getattr(self._client, resource), verb)(
            params, metadata=metadata
        )

        _LOGGER.debug(
            f"space_connector.dispatch :: response_or_iterator: {response_or_iterator}"
        )

        if self._return_type == "dict":
            if isinstance(response_or_iterator, types.GeneratorType):
                return self._generate_response(response_or_iterator)
            else:
                return self._change_message(response_or_iterator)
        else:
            return response_or_iterator

    def _verify(self) -> None:
        if self._service:
            if not isinstance(self._endpoints, dict):
                raise ERROR_CONNECTOR_CONFIGURATION(connector="SpaceConnector")

            if self._service not in self._endpoints:
                raise ERROR_CONNECTOR_LOAD(
                    connector="SpaceConnector",
                    reason=f"{self._service} endpoint is undefined.",
                )

        elif not self._endpoint:
            raise ERROR_CONNECTOR_LOAD(
                connector="SpaceConnector", reason="service or endpoint is required."
            )

    def _init_client(self) -> None:
        endpoint = self._get_endpoint()
        e = parse_grpc_endpoint(endpoint)
        self._client: GRPCClient = pygrpc.client(
            endpoint=e["endpoint"],
            ssl_enabled=e["ssl_enabled"],
            max_message_length=1024 * 1024 * 256,
        )

    @staticmethod
    def _change_message(message) -> dict:
        return MessageToDict(message, preserving_proto_field_name=True)

    def _generate_response(self, response_iterator):
        for response in response_iterator:
            yield self._change_message(response)

    def _get_connection_metadata(
        self, token: str = None, x_domain_id: str = None, x_workspace_id: str = None
    ) -> List[Tuple]:
        metadata = []

        if token:
            metadata.append(("token", token))
        elif self._token:
            metadata.append(("token", self._token))
        elif token := self.transaction.meta.get("token"):
            metadata.append(("token", token))

        if x_domain_id:
            metadata.append(("x_domain_id", x_domain_id))

        if x_workspace_id:
            metadata.append(("x_workspace_id", x_workspace_id))

        carrier = {}
        TraceContextTextMapPropagator().inject(carrier)

        if traceparent := carrier.get("traceparent"):
            metadata.append(("traceparent", traceparent))

        return metadata

    def _parse_method(self, method: str) -> Tuple[str, str]:
        try:
            resource, verb = method.split(".")
        except Exception:
            raise ERROR_CONNECTOR(
                connector="SpaceConnector",
                reason=f"Method is invalid. (endpoint = {self._get_endpoint()}, method = {method})",
            )

        return resource, verb

    def _check_method(self, resource: str, verb: str) -> None:
        supported_verb = self._client.api_resources.get(resource)

        if supported_verb is None or verb not in supported_verb:
            raise ERROR_CONNECTOR(
                connector="SpaceConnector",
                reason=f"Method not supported. "
                f"(endpoint = {self._get_endpoint()}, method = {resource}.{verb})",
            )

    def _get_endpoint(self) -> str:
        return self._endpoint or self._endpoints[self._service]



import logging
import grpc
from concurrent import futures
from typing import List, Union, Type
from grpc_reflection.v1alpha import reflection
from spaceone.core import config
from spaceone.core.pygrpc.api import BaseAPI

_LOGGER = logging.getLogger(__name__)


class _ServerInterceptor(grpc.ServerInterceptor):
    _SKIP_METHODS = "/grpc.reflection.v1alpha.ServerReflection/ServerReflectionInfo"

    def _check_skip_method(self, method):
        is_skip = False

        if method in self._SKIP_METHODS:
            is_skip = True

        return is_skip

    def intercept_service(self, continuation, handler_call_details):
        # is_skip = self._check_skip_method(handler_call_details.method)
        #
        # if not is_skip:
        #     pass

        response = continuation(handler_call_details)

        # if not is_skip:
        #     pass

        return response


class GRPCServer(object):
    def __init__(self):
        conf = config.get_global()
        self._service = conf["SERVICE"]
        self._port = conf["PORT"]
        self._max_workers = conf["MAX_WORKERS"]
        self._service_names = []

        server_interceptor = _ServerInterceptor()
        self._server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=conf["MAX_WORKERS"]),
            interceptors=(server_interceptor,),
        )

    @property
    def server(self) -> grpc.server:
        return self._server

    @property
    def service_names(self) -> List[str]:
        return self._service_names

    def add_service(self, servicer_cls: Union[Type[BaseAPI], Type[object]]):
        servicer = servicer_cls()
        getattr(servicer.pb2_grpc_module, f"add_{servicer.name}Servicer_to_server")(
            servicer, self.server
        )
        self.service_names.append(servicer.service_name)

    def run(self):
        service_names_str = "\n\t - ".join(self.service_names)
        _LOGGER.debug(f"Loaded Services: \n\t - {service_names_str}")
        reflection.enable_server_reflection(self.service_names, self.server)

        self.server.add_insecure_port(f"[::]:{self._port}")
        _LOGGER.info(
            f"Start gRPC Server ({self._service}): "
            f"port={self._port}, max_workers={self._max_workers}"
        )
        self.server.start()
        self.server.wait_for_termination()


def _get_grpc_app() -> GRPCServer:
    package: str = config.get_package()
    app_path: str = config.get_global("GRPC_APP_PATH")
    app_path = app_path.format(package=package)
    module_path, app_name = app_path.split(":")

    try:
        app_module = __import__(module_path, fromlist=[app_name])

        if not hasattr(app_module, "app"):
            raise ImportError(f"App is not defined. (app_path = {package}.{app_path})")

        return getattr(app_module, "app")
    except Exception as e:
        raise ImportError(f"Cannot import app: {e}")


def _import_module(module_path, servicer_name):
    module = None
    try:
        module = __import__(module_path, fromlist=[servicer_name])
    except Exception as e:
        _LOGGER.warning(
            f"[_import_module] Cannot import grpc servicer module. (reason = {e})",
            exc_info=True,
        )

    return module


def add_extension_services(app):
    ext_proto_conf = config.get_global("GRPC_EXTENSION_SERVICERS", {})
    for module_path, servicer_names in ext_proto_conf.items():
        for servicer_name in servicer_names:
            if api_module := _import_module(module_path, servicer_name):
                if hasattr(api_module, servicer_name):
                    servicer_cls = getattr(api_module, servicer_name)

                    app.add_service(servicer_cls)
                else:
                    _LOGGER.warning(
                        f"[_add_services] Failed to add service. "
                        f"(module_path={module_path}, servicer_name={servicer_name})"
                    )

    return app


def serve():
    app = _get_grpc_app()
    app = add_extension_services(app)
    app.run()



from typing import Tuple, List


class BaseModel(object):

    @classmethod
    def init(cls):
        """
        Args:
        Returns:
            None
        """
        raise NotImplementedError('model.init not implemented!')

    @classmethod
    def create(cls, data: dict):
        """
        Args:
            data (dict)
        Returns:
            model_vo (object)
        """
        raise NotImplementedError('model.create not implemented!')

    def update(self, data: dict):
        """
        Args:
            data (dict)
        Returns:
            model_vo (object)
        """
        raise NotImplementedError('model.update not implemented!')

    def delete(self):
        """
        Args:
        Returns:
            None
        """
        raise NotImplementedError('model.delete not implemented!')

    def terminate(self):
        """
        Args:
        Returns:
            None
        """
        raise NotImplementedError('model.terminate not implemented!')

    def increment(self, key: str, amount: int = 1) -> object:
        """
        Args:
            key (str)
            amount (int)
        Returns:
            model_vo (object)
        """
        raise NotImplementedError('model.increment not implemented!')

    def decrement(self, key: str, amount: int = 1):
        """
        Args:
            key (str)
            amount (int)
        Returns:
            model_vo (object)
        """
        raise NotImplementedError('model.decrement not implemented!')

    @classmethod
    def get(cls, **conditions):
        """
        Args:
            **conditions (kwargs)
            - key (str): value (any)
        Returns:
            model_vo (object)
        """
        raise NotImplementedError('model.get not implemented!')

    @classmethod
    def filter(cls, **conditions):
        """
        Args:
            **conditions (kwargs)
            - key (str): value (any)
        Returns:
            model_vos (list)
        """
        raise NotImplementedError('model.filter not implemented!')

    def to_dict(self):
        """
        Args:
        Returns:
            model_data (dict)
        """
        raise NotImplementedError('model.to_dict not implemented!')

    @classmethod
    def query(
            cls,
            *args,
            only: list = None,
            exclude: list = None,
            filter: list = None,
            filter_or: list = None,
            sort: dict = None,
            page: dict = None,
            minimal: bool = False,
            count_only: bool = False,
            **kwargs
    ):
        """
        Args:
            *args (list)
            only (list)
            exclude (list)
            filter (list)
            filter_or (list)
            sort (dict)
            page (dict)
            minimal (bool)
            count_only (bool)
            **kwargs (kwargs)

        Returns:
            model_vos (list)
            total_count (int)
        """
        raise NotImplementedError('model.query not implemented!')

    @classmethod
    def analyze(
            cls,
            *args,
            granularity: str = None,
            fields: dict = None,
            select: dict = None,
            group_by: list = None,
            field_group: list = None,
            filter: list = None,
            filter_or: list = None,
            page: dict = None,
            sort: dict = None,
            start: str = None,
            end: str = None,
            date_field: str = 'date',
            date_field_format: str = '%Y-%m-%d',
            **kwargs
    ):
        """
        Args:
            *args (list)
            granularity (str)
            fields (dict)
            select (dict)
            group_by (list)
            field_group (list)
            filter (list)
            filter_or (list)
            page (dict)
            sort (dict)
            start (str)
            end (str)
            date_field (str)
            date_field_format (str)
            **kwargs (kwargs)

        Returns:
            results (list)
            more (bool)
        """
        raise NotImplementedError('model.analyze not implemented!')



import copy
import json
import random
import string
import logging

from multiprocessing import Process

from spaceone.core import queue, config
from spaceone.core.locator import Locator
from spaceone.core.logger import set_logger
from spaceone.core.error import ERROR_TASK_LOCATOR, ERROR_TASK_METHOD

_LOGGER = logging.getLogger(__name__)


def randomString(stringLength=8):
    """Generate a random string of fixed length"""
    letters = string.ascii_lowercase
    return "".join(random.sample(letters, stringLength))


class SingleTask:
    def __init__(self, single_task: dict):
        self.locator = single_task["locator"]
        self.name = single_task["name"]
        self.metadata = single_task["metadata"]
        self.method = single_task["method"]
        self.params = single_task["params"]
        self._locator = Locator()

    def execute(self):
        """Run method"""
        try:
            if self.locator == "SERVICE":
                caller = self._locator.get_service(self.name, self.metadata)
            elif self.locator == "MANAGER":
                caller = self._locator.get_manager(self.name)

        except Exception as e:
            _LOGGER.debug(f"[SingleTask.execute] locator error: {e}")
            raise ERROR_TASK_LOCATOR(locator=self.locator, name=self.name)

        try:
            _LOGGER.debug(
                f"[SingleTask.execute] (REQUEST) => {self.name}.{self.method}"
            )
            method = getattr(caller, self.method)
            resp = method(**self.params)
            _LOGGER.debug(f"[SingleTask.execute] (RESPONSE) => SUCCESS")
            return resp
        except Exception as e:
            _LOGGER.error(
                f"[SingleTask.execute] Fail to execute method ({self.method}): {e}",
                exc_info=True,
            )
            raise ERROR_TASK_METHOD(name=self.name, method=self.method)


class SpaceoneTask:
    def __init__(self, task: dict):
        self.task = task
        self.name = task.get("name", "")
        self.version = task.get("version", "v1")
        self.execution_engine = task.get("executionEngine", "BaseWorker")
        self.stages = task.get("stages")
        self.stop_on_failure = task.get("stop_on_failure", True)

    def execute(self):
        """Run stage"""
        self.tasks = []
        for stage in self.stages:
            try:
                single_task = SingleTask(stage)
                single_task.execute()
            except Exception as e:
                stage_log = copy.deepcopy(stage)
                stage_log["metadata"] = "*****"
                _LOGGER.error(
                    f"[SpaceoneTask] fail to parse {stage_log}, {e}", exc_info=True
                )
                if self.stop_on_failure:
                    _LOGGER.debug(
                        f"[SpaceoneTask] stop task, since stop on failure is enabled"
                    )
                    break


class BaseWorker(Process):
    def __init__(self, queue, **kwargs):
        self._name_ = "worker-%s" % randomString()
        self.queue = queue
        _LOGGER.debug(f"[BaseWorker] BaseWorker name  : {self._name_}")
        _LOGGER.debug(f"[BaseWorker] BaseWorker queue : {self.queue}")

        self.global_config = config.get_global()
        super().__init__()

    def run(self):
        """Infinite Loop"""
        config.set_global_force(**self.global_config)

        # Enable logging configuration
        set_logger()

        while True:
            # Read from Queue
            binary_task = queue.get(self.queue)
            try:
                json_task = json.loads(binary_task.decode())
                task = SpaceoneTask(json_task)
                # Run task
                task.execute()

            except Exception as e:
                _LOGGER.error(
                    f"[{self._name_}] failed to decode task: {binary_task}, {e}"
                )
                continue



import logging
from enum import Enum

from spaceone.core import config
from grpc_health.v1.health import HealthServicer, SERVICE_NAME
from grpc_health.v1 import health_pb2, health_pb2_grpc

_LOGGER = logging.getLogger(__name__)


class GRPCHealth(HealthServicer):

    def __init__(self, experimental_non_blocking=True,
                 experimental_thread_pool=None):
        super().__init__(experimental_non_blocking, experimental_thread_pool)
        self.health_mgr = HealthManager()
        self.health_mgr.add_health_update(self)
        self.health_mgr.check()

    @property
    def name(self):
        return 'Health'

    @property
    def pb2_grpc_module(self):
        return health_pb2_grpc

    @property
    def service_name(self):
        return SERVICE_NAME

    def Check(self, request, context):
        try:
            status = self.health_mgr.check()
            status = status.value
            self.update_status(status)

        except Exception as e:
            _LOGGER.error(f'[Check] Health Check Error: {e}')
            status = 'UNKNOWN'

        return health_pb2.HealthCheckResponse(status=status)

    def update_status(self, status):
        service_name = config.get_service()
        self.set(service_name, status)


class HealthManager(object):
    _checkers = []

    class Status(Enum):
        UNKNOWN = 'UNKNOWN'
        """When your application's status is indeterminable."""

        SERVING = 'SERVING'
        """When your application is ready."""

        NOT_SERVING = 'NOT_SERVING'
        """When your application is not ready."""

    def check(self):
        status = self.Status.SERVING
        return status

    def add_health_update(self, obj):
        self._checkers.append(obj)

    def update_status(self, status):
        for obj in self._checkers:
            obj.update_status(status.value)


import copy
import logging
import consul

from spaceone.core import utils
from spaceone.core.config import default_conf

_REMOTE_URL = []
_GLOBAL = {}
_LOGGER = logging.getLogger(__name__)


def init_conf(
    package: str,
    port: int = None,
    worker: int = None,
    host: str = None,
    grpc_app_path: str = None,
    rest_app_path: str = None,
    plugin_app_path: str = None,
):
    set_default_conf()

    _GLOBAL["PACKAGE"] = package
    _GLOBAL["SERVICE"] = package.rsplit(".", 1)[-1:][0]

    if host:
        _GLOBAL["HOST"] = host

    if port:
        _GLOBAL["PORT"] = port

    if worker:
        _GLOBAL["MAX_WORKERS"] = worker

    if grpc_app_path:
        _GLOBAL["GRPC_APP_PATH"] = grpc_app_path

    if rest_app_path:
        _GLOBAL["REST_APP_PATH"] = rest_app_path

    if plugin_app_path:
        _GLOBAL["PLUGIN_APP_PATH"] = plugin_app_path


def set_default_conf():
    for key, value in vars(default_conf).items():
        if not key.startswith("__"):
            _GLOBAL[key] = value


def get_package():
    return _GLOBAL["PACKAGE"]


def get_service():
    return _GLOBAL["SERVICE"]


def get_connector(name):
    return _GLOBAL.get("CONNECTORS", {}).get(name, {})


def set_service_config(global_conf_path: str = None):
    """
    Get config from service
    """

    package = _GLOBAL["PACKAGE"]
    if package is None:
        raise ValueError(f"Package is undefined.")

    global_conf_path = global_conf_path or _GLOBAL["GLOBAL_CONF_PATH"]
    global_conf_path = global_conf_path.format(package=package)

    module_path, fromlist = global_conf_path.split(":")
    global_module = __import__(module_path, fromlist=[fromlist])

    for key, value in vars(global_module).items():
        if not key.startswith("__"):
            if key in _GLOBAL:
                if isinstance(value, dict):
                    _GLOBAL[key] = utils.deep_merge(value, _GLOBAL[key])
                else:
                    _GLOBAL[key] = value
            else:
                _GLOBAL[key] = value


def get_global(key=None, default=None):
    if key:
        return copy.deepcopy(_GLOBAL.get(key, default))
    else:
        return copy.deepcopy(_GLOBAL)


def set_global(**config):
    global_conf = get_global()

    for key, value in config.items():
        if key in global_conf:
            if (
                not isinstance(value, type(global_conf[key]))
                and global_conf[key] is not None
            ):
                value_type_name = type(global_conf[key]).__name__
                raise ValueError(
                    f"Value type is invalid. (GLOBAL.{key} = {value_type_name})"
                )

            if isinstance(value, dict):
                global_conf[key] = utils.deep_merge(value, global_conf[key])
            else:
                global_conf[key] = value

    _GLOBAL.update(global_conf)


def set_global_force(**config):
    for key, value in config.items():
        _GLOBAL[key] = value


def set_file_conf(config_yml: str):
    file_conf: dict = utils.load_yaml_from_file(config_yml)
    global_conf: dict = file_conf.get("GLOBAL", {})
    set_global(**global_conf)

    import_conf: list = file_conf.get("IMPORT", [])
    if isinstance(import_conf, list):
        for uri in import_conf:
            import_remote_conf(uri)

    # DEPRECATED: REMOTE_URL setting changed to IMPORT
    import_conf: list = file_conf.get("REMOTE_URL", [])
    if isinstance(import_conf, list):
        for uri in import_conf:
            import_remote_conf(uri)


def import_remote_conf(uri):
    endpoint = utils.parse_endpoint(uri)
    scheme = endpoint.get("scheme")

    remote_conf = None

    if scheme == "file":
        remote_conf = utils.load_yaml_from_file(endpoint["path"])

    elif scheme in ["http", "https"]:
        remote_conf = utils.load_yaml_from_url(uri)

    elif scheme == "consul":
        remote_conf = load_consul_config(endpoint)

    if isinstance(remote_conf, dict):
        set_global(**remote_conf)


def load_consul_config(endpoint):
    hostname = endpoint.get("hostname")
    port = endpoint.get("port")
    key = endpoint.get("path", "")[1:]

    try:
        conf = {}
        if hostname:
            conf["host"] = hostname

        if port:
            conf["port"] = port

        c = consul.Consul(**conf)
        index, data = c.kv.get(key)
        if data:
            json_str = data["Value"].decode("utf-8")
            return utils.load_json(json_str)

        return {}
    except Exception as e:
        raise Exception(f"Consul Call Error: {e}")



import logging
import json


class ParameterFilter(logging.Filter):
    def filter(self, record):
        params = getattr(record, "parameter", None)

        if params is None:
            record.params = ""
        else:
            record.params = json.dumps(params)

        return True


class ParameterLogFilter(logging.Filter):
    def filter(self, record):
        params = getattr(record, "parameter", None)

        if params is None or params == "":
            record.params_log = "{}"
        else:
            record.params_log = json.dumps(params)

        return True



import bson
import re
import logging
import certifi
import copy
from datetime import datetime, date
from dateutil.relativedelta import relativedelta
from functools import reduce, partial
from mongoengine import (
    EmbeddedDocumentField,
    EmbeddedDocument,
    Document,
    QuerySet,
    register_connection,
)
from mongoengine.fields import DateField, DateTimeField, ComplexDateTimeField
from pymongo import ReadPreference
from mongoengine.errors import *
from spaceone.core import config
from spaceone.core import utils
from spaceone.core.error import *
from spaceone.core.model.base_model import BaseModel
from spaceone.core.model.mongo_model.filter_operator import FILTER_OPERATORS
from spaceone.core.model.mongo_model.stat_operator import (
    STAT_GROUP_OPERATORS,
    STAT_PROJECT_OPERATORS,
)

_REFERENCE_ERROR_FORMAT = r"Could not delete document \((\w+)\.\w+ refers to it\)"
_MONGO_INIT_MODELS = []

_LOGGER = logging.getLogger(__name__)


def _raise_reference_error(class_name, message):
    m = re.findall(_REFERENCE_ERROR_FORMAT, message)
    if len(m) > 0:
        raise ERROR_EXIST_RESOURCE(parent=class_name, child=m[0])


class MongoCustomQuerySet(QuerySet):
    def last(self):
        return self.order_by("-id").first()

    def update(self, *args, **kwargs):
        if len(args) > 0 and isinstance(args[0], dict):
            kwargs.update(args[0])
        super().update(**kwargs)

    def increment(self, key, amount=1):
        key = key.replace(".", "__")
        inc_data = {f"inc__{key}": amount}

        super().update(**inc_data)

    def decrement(self, key, amount=1):
        key = key.replace(".", "__")
        dec_data = {f"dec__{key}": amount}

        super().update(**dec_data)

    def set_data(self, key, data):
        key = key.replace(".", "__")
        set_data = {f"set__{key}": data}

        super().update(**set_data)

    def unset_data(self, *keys):
        unset_data = {}

        for key in keys:
            key = key.replace(".", "__")
            unset_data[f"unset__{key}"] = 1

        super().update(**unset_data)

    def append(self, key, data):
        key = key.replace(".", "__")
        append_data = {f"push__{key}": data}

        super().update(**append_data)

    def remove(self, key, data):
        key = key.replace(".", "__")
        remove_data = {f"pull__{key}": data}
        super().update(**remove_data)


class MongoModel(Document, BaseModel):
    auto_create_index = True
    meta = {
        "abstract": True,
        "queryset_class": MongoCustomQuerySet,
        "auto_create_index": False,
    }

    @classmethod
    def init(cls, create_index: bool = True) -> None:
        global_conf = config.get_global()
        databases = global_conf.get("DATABASES", {})
        db_name_prefix = global_conf.get("DATABASE_NAME_PREFIX", "")

        if not global_conf.get("MOCK_MODE", False):
            for alias, db_conf in databases.items():
                is_connect = cls._connect(alias, db_conf, db_name_prefix)
                if is_connect:
                    package_path = config.get_package()
                    model_path = config.get_global("DATABASE_MODEL_PATH", "model")
                    __import__(f"{package_path}.{model_path}", fromlist=["*"])

                    for model in cls.__subclasses__():
                        if create_index:
                            model._create_index()
                        model._load_default_meta()

    @classmethod
    def _connect(cls, alias: str, db_conf: dict, db_name_prefix: str) -> bool:
        is_connect = False
        db_conf = copy.deepcopy(db_conf)
        engine = db_conf.get("engine")
        host = db_conf.get("host")
        db = db_conf.get("db")
        username = db_conf.get("username")

        if engine is None:
            raise ERROR_DB_ENGINE_UNDEFINE(alias=alias)
        del db_conf["engine"]

        if engine == "MongoModel":
            if host and db and username:
                if "read_preference" in db_conf:
                    read_preference = getattr(
                        ReadPreference, db_conf["read_preference"], None
                    )
                    if read_preference:
                        db_conf["read_preference"] = read_preference
                    else:
                        del db_conf["read_preference"]

                db_name: str = db_conf.get("db", "")
                db_conf["db"] = f"{db_name_prefix}{db_name}"

                host: str = str(db_conf.get("host", "")).strip()
                if host.startswith("mongodb+srv://"):
                    db_conf["tlsCAFile"] = certifi.where()

                register_connection(alias, **db_conf)
                is_connect = True
                _LOGGER.debug(f"Create MongoDB Connection: {alias}")

        return is_connect

    @classmethod
    def _load_default_meta(cls):
        cls._meta["datetime_fields"] = []
        for name, field in cls._fields.items():
            if isinstance(field, DateField):
                cls._meta["datetime_fields"].append(name)
            elif isinstance(field, DateTimeField):
                cls._meta["datetime_fields"].append(name)
            elif isinstance(field, ComplexDateTimeField):
                cls._meta["datetime_fields"].append(name)

    @classmethod
    def _create_index(cls) -> None:
        if cls.auto_create_index:
            indexes = cls._meta.get("indexes", [])
            unique_fields = cls._get_unique_fields()

            if len(indexes) > 0 or len(unique_fields) > 0:
                total_index_count = len(indexes) + len(unique_fields)

                _LOGGER.debug(
                    f"Create MongoDB Indexes ({cls.__name__} Model: {total_index_count} Indexes)"
                )

                for unique_field in unique_fields:
                    try:
                        cls.create_index({"fields": unique_field, "unique": True})

                    except Exception as e:
                        _LOGGER.error(f"Unique Index Creation Failure: {e}")

                for index in indexes:
                    try:
                        cls.create_index(index)

                    except Exception as e:
                        if e.code != 85:  # 85: IndexOptionsConflict
                            _LOGGER.error(f"Index Creation Failure: {e}")

    @classmethod
    def _get_unique_fields(cls):
        unique_fields = []
        for name, field in cls._fields.items():
            if field.unique:
                if isinstance(field.unique_with, str):
                    unique_fields.append([field.name, field.unique_with])
                elif isinstance(field.unique_with, list):
                    unique_fields.append([field.name] + field.unique_with)
                else:
                    unique_fields.append([field.name])

        return unique_fields

    @classmethod
    def _trim_value(cls, value: any) -> any:
        if isinstance(value, str):
            return value.strip()
        elif isinstance(value, dict):
            changed_value = {}
            for k, v in value.items():
                if isinstance(k, str):
                    k = k.strip()
                if isinstance(v, str):
                    v = v.strip()
                changed_value[k] = cls._trim_value(v)

            return changed_value
        else:
            return value

    @classmethod
    def create(cls, data):
        create_data = {}

        for name, field in cls._fields.items():
            if name in data:
                create_data[name] = data[name]
            else:
                generate_id = getattr(field, "generate_id", None)
                if generate_id:
                    create_data[name] = utils.generate_id(generate_id)

                if getattr(field, "auto_now", False):
                    create_data[name] = datetime.utcnow()
                elif getattr(field, "auto_now_add", False):
                    create_data[name] = datetime.utcnow()

        for unique_field in cls._get_unique_fields():
            conditions = {}
            for f in unique_field:
                conditions[f] = data.get(f)
            vos = cls.filter(**conditions)
            if vos.count() > 0:
                raise ERROR_SAVE_UNIQUE_VALUES(keys=unique_field)

        for key, value in create_data.items():
            create_data[key] = cls._trim_value(value)

        try:
            new_vo = cls(**create_data).save()
        except Exception as e:
            raise ERROR_DB_QUERY(reason=e)

        return new_vo

    def update(self, data):
        updatable_fields = self._meta.get(
            "updatable_fields",
            list(
                filter(
                    lambda x: x != self._meta.get("id_field", "id"), self._fields.keys()
                )
            ),
        )

        for name, field in self._fields.items():
            if getattr(field, "auto_now", False):
                if name not in data.keys():
                    data[name] = datetime.utcnow()

        for unique_field in self._get_unique_fields():
            conditions = {"pk__ne": self.pk}
            for f in unique_field:
                conditions[f] = data.get(f)

            vos = self.filter(**conditions)
            if vos.count() > 0:
                raise ERROR_SAVE_UNIQUE_VALUES(keys=unique_field)

        for key in list(data.keys()):
            if key not in updatable_fields:
                del data[key]

        if data != {}:
            for key, value in data.items():
                data[key] = self._trim_value(value)

            try:
                super().update(**data)
                self.reload()
            except Exception as e:
                raise ERROR_DB_QUERY(reason=e)

        return self

    def delete(self, *args):
        try:
            super().delete(*args)
        except OperationError as e:
            _raise_reference_error(self.__class__.__name__, str(e))
            raise ERROR_DB_QUERY(reason=e)
        except Exception as e:
            raise ERROR_DB_QUERY(reason=e)

    def terminate(self):
        super().delete()

    def increment(self, key, amount=1):
        key = key.replace(".", "__")
        inc_data = {f"inc__{key}": amount}

        super().update(**inc_data)
        self.reload()
        return self

    def decrement(self, key, amount=1):
        key = key.replace(".", "__")
        dec_data = {f"dec__{key}": amount}

        super().update(**dec_data)
        self.reload()
        return self

    def set_data(self, key, data):
        key = key.replace(".", "__")
        set_data = {f"set__{key}": data}

        super().update(**set_data)
        self.reload()
        return self

    def unset_data(self, *keys):
        unset_data = {}

        for key in keys:
            key = key.replace(".", "__")
            unset_data[f"unset__{key}"] = 1

        super().update(**unset_data)
        self.reload()
        return self

    def append(self, key, data) -> Document:
        key = key.replace(".", "__")
        append_data = {}

        field = getattr(self._fields.get(key, {}), "field", None)
        if field and isinstance(field, EmbeddedDocumentField):
            reference_model = field.document_type_obj
            append_data[f"push__{key}"] = reference_model(**data)
        else:
            append_data[f"push__{key}"] = data

        super().update(**append_data)
        self.reload()
        return self

    def remove(self, key, data):
        key = key.replace(".", "__")
        remove_data = {f"pull__{key}": data}
        super().update(**remove_data)
        self.reload()
        return self

    @classmethod
    def get(cls, only=None, **conditions):
        vos = cls.filter(**conditions)

        if vos.count() == 0:
            keys = tuple(conditions.keys())
            values = tuple(conditions.values())

            if len(keys) == 1:
                raise ERROR_NOT_FOUND(key=keys[0], value=values[0])
            else:
                raise ERROR_NOT_FOUND(key=keys, value=values)

        if only:
            only = cls._remove_duplicate_only_keys(only)
            vos = vos.only(*only)

        return vos.first()

    @classmethod
    def filter(cls, **conditions):
        change_conditions = {}
        for key, value in conditions.items():
            if isinstance(value, list):
                change_conditions[f"{key}__in"] = value
            else:
                change_conditions[key] = value

        return cls.objects.filter(**change_conditions)

    def to_dict(self) -> dict:
        return dict(self.to_mongo())

    @classmethod
    def _get_target_objects(cls, target):
        if target:
            read_preference = getattr(ReadPreference, target, None)
            if read_preference:
                return cls.objects.read_preference(read_preference)

        return cls.objects

    @staticmethod
    def _check_operator_value(is_multiple, operator, value, condition):
        if is_multiple:
            if not isinstance(value, list):
                raise ERROR_OPERATOR_LIST_VALUE_TYPE(
                    operator=operator, condition=condition
                )

        else:
            if isinstance(value, list):
                raise ERROR_OPERATOR_VALUE_TYPE(operator=operator, condition=condition)

    @classmethod
    def _check_reference_field(cls, key):
        ref_keys = cls._meta.get("reference_query_keys", {}).keys()
        if key in ref_keys:
            return False
        else:
            return True

    @classmethod
    def _get_reference_model(cls, key):
        for ref_key, ref_conf in cls._meta.get("reference_query_keys", {}).items():
            if key.startswith(ref_key) and key[len(ref_key)] == ".":
                if isinstance(ref_conf, dict):
                    ref_model = ref_conf.get("model")
                    foreign_key = ref_conf.get("foreign_key")
                else:
                    ref_model = ref_conf
                    foreign_key = None

                ref_query_key = key.replace(f"{ref_key}.", "")
                if ref_model == "self":
                    ref_model = cls

                return ref_model, ref_key, ref_query_key, foreign_key

        return None, None, None, None

    @classmethod
    def _change_reference_condition(cls, key, value, operator, reference_filter=None):
        ref_model, ref_key, ref_query_key, foreign_key = cls._get_reference_model(key)
        if ref_model:
            if value is None:
                return ref_key, value, operator
            else:
                if operator == "not":
                    _filter = [{"k": ref_query_key, "v": value, "o": "eq"}]
                elif operator == "not_in":
                    _filter = [{"k": ref_query_key, "v": value, "o": "in"}]
                else:
                    _filter = [{"k": ref_query_key, "v": value, "o": operator}]
                if reference_filter:
                    for key, value in reference_filter.items():
                        if value:
                            _filter.append({"k": key, "v": value, "o": "eq"})

                ref_vos, total_count = ref_model.query(filter=_filter)

                if foreign_key:
                    ref_values = []
                    for ref_vo in ref_vos:
                        ref_value = getattr(ref_vo, foreign_key)
                        if ref_value:
                            ref_values.append(ref_value)
                else:
                    ref_values = list(ref_vos)

                if operator in ["not", "not_in"]:
                    return ref_key, ref_values, "not_in"
                else:
                    return ref_key, ref_values, "in"

        else:
            return key, value, operator

    @classmethod
    def _make_condition(cls, condition, reference_filter=None):
        key = condition.get("key", condition.get("k"))
        value = condition.get("value", condition.get("v"))
        operator = condition.get("operator", condition.get("o"))
        change_query_keys = cls._meta.get("change_query_keys", {})

        if operator not in FILTER_OPERATORS:
            raise ERROR_DB_QUERY(
                reason=f"Filter operator is not supported. (operator = "
                f"{FILTER_OPERATORS.keys()})"
            )

        resolver, mongo_operator, is_multiple = FILTER_OPERATORS.get(operator)

        cls._check_operator_value(is_multiple, operator, value, condition)

        if key and operator:
            if key in change_query_keys:
                key = change_query_keys[key]

            if operator not in ["regex", "regex_in"]:
                if cls._check_reference_field(key):
                    key, value, operator = cls._change_reference_condition(
                        key, value, operator, reference_filter
                    )

                    resolver, mongo_operator, is_multiple = FILTER_OPERATORS[operator]

                key = key.replace(".", "__")

            return resolver(key, value, mongo_operator, is_multiple)
        else:
            raise ERROR_DB_QUERY(
                reason="Filter condition should have key, value and operator."
            )

    @classmethod
    def _make_filter(cls, filter, filter_or, reference_filter):
        _filter = None
        _filter_or = None

        if len(filter) > 0:
            _filter = reduce(
                lambda x, y: x & y,
                map(
                    partial(cls._make_condition, reference_filter=reference_filter),
                    filter,
                ),
            )

        if len(filter_or) > 0:
            _filter_or = reduce(
                lambda x, y: x | y,
                map(
                    partial(cls._make_condition, reference_filter=reference_filter),
                    filter_or,
                ),
            )

        if _filter and _filter_or:
            _filter = _filter & _filter_or
        else:
            _filter = _filter or _filter_or

        return _filter

    @classmethod
    def _remove_duplicate_only_keys(cls, only):
        changed_only = []
        duplicated_only = []
        for key in only:
            exists = False

            for changed_key in changed_only:
                if key == changed_key or key.startswith(f"{changed_key}."):
                    exists = True
                elif changed_key.startswith(f"{key}."):
                    duplicated_only.append(changed_key)

            if exists is False:
                changed_only.append(key)

        if len(duplicated_only) > 0:
            changed_only = list(set(changed_only) - set(duplicated_only))

        return changed_only

    @staticmethod
    def _make_unwind_project_stage(only: list):
        project_fields = []
        for key in only:
            project_fields.append(
                {
                    "key": key,
                    "name": key,
                }
            )

        return {
            "project": {
                "exclude_keys": False,
                "only_keys": False,
                "fields": project_fields,
            }
        }

    @classmethod
    def _stat_with_unwind(
        cls,
        unwind: list,
        only: list = None,
        filter: list = None,
        filter_or: list = None,
        sort: list = None,
        page: dict = None,
        target: str = None,
    ):
        if only is None:
            raise ERROR_DB_QUERY(reason="unwind option requires only option.")

        if not isinstance(unwind, dict):
            raise ERROR_DB_QUERY(reason="unwind option should be dict type.")

        if "path" not in unwind:
            raise ERROR_DB_QUERY(reason="unwind option should have path key.")

        unwind_path = unwind["path"]
        aggregate = [{"unwind": unwind}]

        # Add project stage
        project_fields = []
        for key in only:
            project_fields.append(
                {
                    "key": key,
                    "name": key,
                }
            )

        aggregate.append(
            {
                "project": {
                    "exclude_keys": True,
                    "only_keys": True,
                    "fields": project_fields,
                }
            }
        )

        # Add sort stage
        if sort:
            aggregate.append({"sort": sort})

        response = cls.stat(
            aggregate=aggregate,
            filter=filter,
            filter_or=filter_or,
            page=page,
            tageet=target,
            allow_disk_use=True,
        )

        try:
            vos = []
            total_count = response.get("total_count", 0)
            for result in response.get("results", []):
                unwind_data = utils.get_dict_value(result, unwind_path)
                result = utils.change_dict_value(result, unwind_path, [unwind_data])

                vo = cls(**result)
                vos.append(vo)
        except Exception as e:
            raise ERROR_DB_QUERY(reason=f"Failed to convert unwind result: {e}")

        return vos, total_count

    @classmethod
    def query(
        cls,
        *args,
        only=None,
        exclude=None,
        filter=None,
        filter_or=None,
        sort=None,
        page=None,
        minimal=False,
        include_count=True,
        count_only=False,
        unwind=None,
        reference_filter=None,
        target=None,
        hint=None,
        **kwargs,
    ):
        filter = filter or []
        filter_or = filter_or or []
        sort = sort or []
        page = page or {}

        if unwind:
            return cls._stat_with_unwind(
                unwind, only, filter, filter_or, sort, page, target
            )

        else:
            _order_by = []
            minimal_fields = cls._meta.get("minimal_fields")

            _filter = cls._make_filter(filter, filter_or, reference_filter)

            for sort_option in sort:
                if sort_option.get("desc", False):
                    _order_by.append(f'-{sort_option["key"]}')
                else:
                    _order_by.append(f'{sort_option["key"]}')

            try:
                vos = cls._get_target_objects(target).filter(_filter)

                if hint:
                    vos = vos.hint(hint)

                if len(_order_by) > 0:
                    vos = vos.order_by(*_order_by)

                if only:
                    if len(_order_by) > 0:
                        ordering = _order_by
                    else:
                        ordering = cls._meta.get("ordering")

                    for key in ordering:
                        if key.startswith("+") or key.startswith("-"):
                            key = key[1:]
                        if key not in only:
                            only.append(key)

                    only = cls._remove_duplicate_only_keys(only)
                    vos = vos.only(*only)

                if exclude:
                    vos = vos.exclude(*exclude)

                if minimal and minimal_fields:
                    vos = vos.only(*minimal_fields)

                if include_count:
                    total_count = vos.count()
                else:
                    total_count = 0

                if count_only:
                    vos = []

                else:
                    if "limit" in page and page["limit"] > 0:
                        start = page.get("start", 1)
                        if start < 1:
                            start = 1

                        vos = vos[start - 1 : start + page["limit"] - 1]

                return vos, total_count

            except Exception as e:
                raise ERROR_DB_QUERY(reason=e)

    @classmethod
    def _check_well_known_type(cls, value):
        if isinstance(value, bson.objectid.ObjectId):
            return str(value)
        elif isinstance(value, Document):
            return str(value.id)
        elif isinstance(value, EmbeddedDocument):
            return dict(value.to_mongo())
        else:
            return value

    @classmethod
    def _make_aggregate_values(cls, cursor):
        values = []
        for row in cursor:
            data = {}
            for key, value in row.items():
                if key == "_id" and isinstance(row[key], dict):
                    for group_key, group_value in row[key].items():
                        data[group_key] = cls._check_well_known_type(group_value)
                else:
                    data[key] = cls._check_well_known_type(value)

            values.append(data)

        return values

    @classmethod
    def _make_distinct_values(cls, values):
        changed_values = []
        for value in values:
            changed_values.append(cls._check_well_known_type(value))

        return changed_values

    @classmethod
    def _make_sub_conditions(cls, sub_conditions, _before_group_keys):
        and_sub_conditions = []

        for sub_condition in sub_conditions:
            key = sub_condition.get("key", sub_condition.get("k"))
            value = sub_condition.get("value", sub_condition.get("v"))
            operator = sub_condition.get("operator", sub_condition.get("o"))

            if key is None:
                raise ERROR_DB_QUERY(
                    reason=f"'aggregate.group.fields.conditions.key' condition requires a key: {sub_condition}"
                )

            if value is None:
                raise ERROR_DB_QUERY(
                    reason=f"'aggregate.group.fields.conditions.value' condition requires a value: {sub_condition}"
                )

            if operator is None:
                raise ERROR_DB_QUERY(
                    reason=f"'aggregate.group.fields.conditions.operator' condition requires a operator: {sub_condition}"
                )

            _SUPPORTED_OPERATOR = ["eq", "not", "gt", "gte", "lt", "lte"]

            if operator not in _SUPPORTED_OPERATOR:
                raise ERROR_DB_QUERY(
                    reason=f"'aggregate.group.fields.conditions.operator' condition's {operator} operator is not "
                    f"supported. (supported_operator = {_SUPPORTED_OPERATOR})"
                )

            if key in _before_group_keys:
                key = f"_id.{key}"

            and_sub_conditions.append({f"${operator}": [f"${key}", value]})

        return {"$and": and_sub_conditions}

    @classmethod
    def _get_group_fields(cls, condition, _before_group_keys):
        key = condition.get("key", condition.get("k"))
        name = condition.get("name", condition.get("n"))
        operator = condition.get("operator", condition.get("o"))
        sub_conditions = condition.get("conditions")
        sub_fields = condition.get("fields") or []
        data_type = condition.get("data_type")

        if operator not in STAT_GROUP_OPERATORS:
            raise ERROR_DB_QUERY(
                reason=f"'aggregate.group.fields' condition's {operator} operator is not supported. "
                f"(supported_operator = {list(STAT_GROUP_OPERATORS.keys())})"
            )

        if name is None:
            raise ERROR_DB_QUERY(
                reason=f"'aggregate.group.fields' condition requires a name: {condition}"
            )

        if key in _before_group_keys:
            key = f"_id.{key}"

        for sub_field in sub_fields:
            f_key = sub_field.get("key", sub_field.get("k"))
            f_name = sub_field.get("name", sub_field.get("n"))

            if f_key is None:
                raise ERROR_DB_QUERY(
                    reason=f"'aggregate.group.fields.fields' condition requires a key: {condition}"
                )

            if f_name is None:
                raise ERROR_DB_QUERY(
                    reason=f"'aggregate.group.fields.fields' condition requires a name: {condition}"
                )

            if f_key in _before_group_keys:
                sub_field["key"] = f"_id.{f_key}"

        if sub_conditions:
            sub_conditions = cls._make_sub_conditions(
                sub_conditions, _before_group_keys
            )

        return key, name, operator, sub_fields, sub_conditions, data_type

    @classmethod
    def _get_group_keys(cls, condition, _before_group_keys):
        key = condition.get("key", condition.get("k"))
        name = condition.get("name", condition.get("n"))
        date_format = condition.get("date_format")

        if key is None:
            raise ERROR_DB_QUERY(
                reason=f"'aggregate.group.keys' condition requires a key: {condition}"
            )

        if name is None:
            raise ERROR_DB_QUERY(
                reason=f"'aggregate.group.keys' condition requires a name: {condition}"
            )

        if key in _before_group_keys:
            key = f"_id.{key}"

        if date_format:
            if date_format == "year":
                rule = {"$year": f"${key}"}
            elif date_format == "month":
                rule = {"$month": f"${key}"}
            elif date_format == "day":
                rule = {"$dayOfMonth": f"${key}"}
            else:
                rule = {"$dateToString": {"format": date_format, "date": f"${key}"}}
        else:
            rule = f"${key}"

        return name, rule

    @classmethod
    def _make_group_rule(cls, options, _before_group_keys):
        _group_keys = []
        _group_rule = {"$group": {"_id": {}}}
        _keys = options.get("keys", [])
        _fields = options.get("fields", [])

        if len(_keys) == 0 and len(_fields) == 0:
            raise ERROR_REQUIRED_PARAMETER(
                key="aggregate.group.keys || aggregate.group.fields"
            )

        for condition in _keys:
            name, rule = cls._get_group_keys(condition, _before_group_keys)
            _group_rule["$group"]["_id"][name] = rule
            _group_keys.append(name)

        for condition in _fields:
            (
                key,
                name,
                operator,
                sub_fields,
                sub_conditions,
                data_type,
            ) = cls._get_group_fields(condition, _before_group_keys)

            rule = STAT_GROUP_OPERATORS[operator](
                condition,
                key,
                operator,
                name,
                data_type,
                sub_conditions,
                sub_fields,
                cls._meta["datetime_fields"],
            )
            _group_rule["$group"].update(rule)

        return _group_rule, _group_keys

    @classmethod
    def _get_project_fields(cls, condition):
        key = condition.get("key", condition.get("k"))
        name = condition.get("name", condition.get("n"))
        operator = condition.get("operator", condition.get("o"))
        fields = condition.get("fields", condition.get("f"))

        if operator and operator not in STAT_PROJECT_OPERATORS:
            raise ERROR_DB_QUERY(
                reason=f"'aggregate.project.fields' condition's {operator} operator is not supported. "
                f"(supported_operator = {list(STAT_PROJECT_OPERATORS.keys())})"
            )

        if name is None:
            raise ERROR_DB_QUERY(
                reason=f"'aggregate.project.fields' condition requires a name: {condition}"
            )

        return key, name, operator, fields

    @classmethod
    def _make_project_rule(cls, options, _group_keys):
        _rules = []
        _fields = options.get("fields", [])
        _exclude_keys = options.get("exclude_keys", False)
        _only_keys = options.get("only_keys", False)
        _project_rule = {"$project": {}}

        if len(_fields) == 0:
            raise ERROR_REQUIRED_PARAMETER(key="aggregate.project.fields")

        for condition in _fields:
            key, name, operator, fields = cls._get_project_fields(condition)

            if operator:
                rule = STAT_PROJECT_OPERATORS[operator](
                    condition, key, operator, name, fields, _group_keys
                )
                _project_rule["$project"].update(rule)
            else:
                if key in _group_keys:
                    key = f"_id.{key}"

                if _only_keys:
                    _project_rule["$project"][name] = 1
                else:
                    _project_rule["$project"][name] = f"${key}"

        if _exclude_keys:
            _project_rule["$project"]["_id"] = 0
            _group_keys = []

        return _project_rule, _group_keys

    @classmethod
    def _make_unwind_rule(cls, options):
        if "path" not in options:
            raise ERROR_REQUIRED_PARAMETER(key="aggregate.unwind.path")

        return {"$unwind": f"${options['path']}"}

    @classmethod
    def _make_count_rule(cls, options):
        if "name" not in options:
            raise ERROR_REQUIRED_PARAMETER(key="aggregate.count.name")

        return {"$count": options["name"]}

    @classmethod
    def _make_sort_rule(cls, options, _group_keys):
        order_by = {}

        for sort_option in options:
            if sort_option["key"] in _group_keys:
                sort_key = f'_id.{sort_option["key"]}'
            else:
                sort_key = sort_option["key"]

            order_by[sort_key] = -1 if sort_option.get("desc", False) else 1

        return {"$sort": order_by}

    @classmethod
    def _make_limit_rule(cls, options):
        return {"$limit": options}

    @classmethod
    def _make_skip_rule(cls, options):
        return {"$skip": options}

    @classmethod
    def _make_match_rule(cls, options):
        match_options = {"$and": []}

        for condition in options.get("filter", []):
            key = condition.get("key", condition.get("k"))
            value = condition.get("value", condition.get("v"))
            operator = condition.get("operator", condition.get("o"))

            if not (key and value and operator):
                raise ERROR_REQUIRED_PARAMETER(key="aggregate.match.filter")

            if operator == "eq":
                match_options["$and"].append({key: value})
            elif operator == "not":
                match_options["$and"].append({key: {"$ne": value}})
            elif operator == "in":
                match_options["$and"].append({key: {"$in": value}})
            elif operator == "not_in":
                match_options["$and"].append({key: {"$nin": value}})
            elif operator == "contain":
                match_options["$and"].append({key: {"$regex": f".*{value}.*"}})
            elif operator == "not_contain":
                match_options["$and"].append(
                    {key: {"$not": {"$regex": f".*{value}.*"}}}
                )
            elif operator == "contain_in":
                match_options["$and"].append(
                    {"$or": [{key: {"$regex": f".*{v}.*"}} for v in value]}
                )
            elif operator == "not_contain_in":
                match_options["$and"].append(
                    {"$nor": [{key: {"$regex": f".*{v}.*"}} for v in value]}
                )
            else:
                raise ERROR_DB_QUERY(reason=f"Unsupported operator: {operator}")

        return {"$match": match_options}

    @classmethod
    def _make_aggregate_rules(cls, aggregate):
        _aggregate_rules = []
        _group_keys = []

        if not isinstance(aggregate, list):
            raise ERROR_INVALID_PARAMETER_TYPE(key="aggregate", type="list")

        for stage in aggregate:
            if "unwind" in stage:
                rule = cls._make_unwind_rule(stage["unwind"])
                _aggregate_rules.append(rule)

                if unwind_filter := stage["unwind"].get("filter"):
                    rule = cls._make_match_rule({"filter": unwind_filter})
                    _aggregate_rules.append(rule)

            elif "group" in stage:
                rule, group_keys = cls._make_group_rule(stage["group"], _group_keys)
                _aggregate_rules.append(rule)
                _group_keys += group_keys
            elif "count" in stage:
                rule = cls._make_count_rule(stage["count"])
                _aggregate_rules.append(rule)
            elif "sort" in stage:
                rule = cls._make_sort_rule(stage["sort"], _group_keys)
                _aggregate_rules.append(rule)
            elif "project" in stage:
                rule, _group_keys = cls._make_project_rule(
                    stage["project"], _group_keys
                )
                _aggregate_rules.append(rule)
            elif "limit" in stage:
                rule = cls._make_limit_rule(stage["limit"])
                _aggregate_rules.append(rule)
            elif "skip" in stage:
                rule = cls._make_skip_rule(stage["skip"])
                _aggregate_rules.append(rule)
            elif "match" in stage:
                rule = cls._make_match_rule(stage["match"])
                _aggregate_rules.append(rule)
            else:
                raise ERROR_REQUIRED_PARAMETER(
                    key="aggregate.unwind or aggregate.group or "
                    "aggregate.count or aggregate.sort or "
                    "aggregate.project or aggregate.limit or "
                    "aggregate.skip"
                )

        return _aggregate_rules

    @classmethod
    def _stat_aggregate(cls, vos, aggregate, page, hint, allow_disk_use, return_type):
        result = {}
        pipeline = []
        _aggregate_rules = cls._make_aggregate_rules(aggregate)

        for rule in _aggregate_rules:
            pipeline.append(rule)

        if "limit" in page and page["limit"] > 0:
            limit = page["limit"]
            start = page.get("start", 1)
            start = 1 if start < 1 else start

            result["total_count"] = 0
            cursor = vos.aggregate(pipeline + [{"$count": "total_count"}])
            for c in cursor:
                result["total_count"] = c["total_count"]
                break

            if start > 1:
                pipeline.append({"$skip": start - 1})

            pipeline.append({"$limit": limit})

        options = {}
        if allow_disk_use:
            _LOGGER.debug(f"[_stat_aggregate] allow_disk_use: {allow_disk_use}")
            options["allowDiskUse"] = True

        if hint:
            options["hint"] = hint

        cursor = vos.aggregate(pipeline, **options)

        if return_type == "cursor":
            return cursor
        else:
            result["results"] = cls._make_aggregate_values(cursor)
            return result

    @classmethod
    def _stat_distinct(cls, vos, distinct, page):
        result = {}

        values = vos.distinct(distinct)

        try:
            values.sort()
        except Exception as e:
            _LOGGER.debug(f"[_stat_distinct] Failed to sort values: {e}")

        if "limit" in page and page["limit"] > 0:
            start = page.get("start", 1)
            if start < 1:
                start = 1

            result["total_count"] = len(values)
            values = values[start - 1 : start + page["limit"] - 1]

        result["results"] = cls._make_distinct_values(values)
        return result

    @classmethod
    def stat(
        cls,
        *args,
        aggregate=None,
        distinct=None,
        filter=None,
        filter_or=None,
        page=None,
        reference_filter=None,
        target="SECONDARY_PREFERRED",
        hint=None,
        allow_disk_use=False,
        return_type="dict",
        **kwargs,
    ):
        filter = filter or []
        filter_or = filter_or or []
        page = page or {}

        if not (aggregate or distinct):
            raise ERROR_REQUIRED_PARAMETER(key="aggregate")

        _filter = cls._make_filter(filter, filter_or, reference_filter)

        try:
            vos = cls._get_target_objects(target).filter(_filter)

            if aggregate:
                return cls._stat_aggregate(
                    vos, aggregate, page, hint, allow_disk_use, return_type
                )

            elif distinct:
                return cls._stat_distinct(vos, distinct, page)

        except Exception as e:
            if not isinstance(e, ERROR_BASE):
                e = ERROR_UNKNOWN(message=str(e))

            raise ERROR_DB_QUERY(reason=e.message)

    @classmethod
    def _check_field_group(cls, field_group):
        for key in field_group:
            if key.startswith("_total_"):
                raise ERROR_INVALID_PARAMETER(
                    key="field_group",
                    reason="Field group keys cannot contain _total_ characters.",
                )

    @classmethod
    def _make_group_keys(cls, group_by, date_field, granularity=None):
        group_keys = []
        for group_option in group_by:
            if isinstance(group_option, dict):
                key = group_option.get("key")
                name = group_option.get("name")

                if not (key and name):
                    raise ERROR_INVALID_PARAMETER(
                        key="group_by",
                        reason="group_by option requires a key and name.",
                    )
            elif isinstance(group_option, str):
                key = group_option
                name = key.rsplit(".", 1)[-1:][0]
            else:
                raise ERROR_INVALID_PARAMETER(
                    key="group_by",
                    reason="group_by option should be dict or str type.",
                )

            group_keys.append({"key": key, "name": name})

        if granularity and granularity in ["DAILY", "MONTHLY", "YEARLY"]:
            group_keys.append({"key": date_field, "name": "date"})

        return group_keys

    @classmethod
    def _make_group_fields(cls, fields):
        group_fields = []
        for name, condition in fields.items():
            cls._check_condition(condition)
            operator = condition["operator"]
            key = condition.get("key")
            fields = condition.get("fields")
            data_type = condition.get("data_type")

            group_field = {"name": name, "operator": operator, "data_type": data_type}

            if operator != "count":
                group_field["key"] = key

            if operator == "push" and fields:
                push_fields = []
                for key, value in fields.items():
                    push_fields.append({"key": value, "name": key})

                group_field["fields"] = push_fields

            group_fields.append(group_field)

        return group_fields

    @classmethod
    def _check_condition(cls, condition):
        key = condition.get("key")
        operator = condition.get("operator")
        fields = condition.get("fields", [])

        if operator is None:
            raise ERROR_REQUIRED_PARAMETER(key="query.fields.operator")

        # Check Operator
        if operator not in ["count", "push"] and key is None:
            raise ERROR_REQUIRED_PARAMETER(key="query.fields.key")

        if operator == "push" and not key and len(fields) == 0:
            raise ERROR_REQUIRED_PARAMETER(key="query.fields.fields")

    @classmethod
    def _make_field_group_keys(cls, group_keys, field_group):
        field_group_keys = []
        for group_option in group_keys:
            key = group_option["name"]
            name = group_option["name"]
            if name not in field_group:
                if name == "date":
                    field_group_keys.append({"key": "date", "name": "date"})
                else:
                    field_group_keys.append({"key": key, "name": name})

        return field_group_keys

    @classmethod
    def _make_field_group_fields(cls, group_fields, field_group):
        field_group_fields = []
        for group_field in group_fields:
            field_name = group_field["name"]
            operator = group_field["operator"]

            if operator in ["sum", "average", "max", "min", "count"]:
                field_group_fields.append(
                    {
                        "key": field_name,
                        "name": f"_total_{field_name}",
                        "operator": "sum" if operator == "count" else operator,
                    }
                )

            field_group_fields.append(
                {
                    "name": field_name,
                    "operator": "push",
                    "fields": cls._make_field_group_push_fields(
                        field_name, field_group
                    ),
                }
            )

        return field_group_fields

    @classmethod
    def _make_field_group_push_fields(cls, field_name, field_group):
        push_fields = [{"name": "value", "key": field_name}]

        for field_group_key in field_group:
            push_fields.append({"name": field_group_key, "key": field_group_key})

        return push_fields

    @classmethod
    def _make_field_group_query(cls, group_keys, group_fields, field_group):
        field_group_query = {
            "group": {
                "keys": cls._make_field_group_keys(group_keys, field_group),
                "fields": cls._make_field_group_fields(group_fields, field_group),
            }
        }

        return [field_group_query]

    @classmethod
    def _make_select_query(cls, select):
        select_query = {"project": {"fields": [], "exclude_keys": True}}

        supported_operator = ["add", "subtract", "multiply", "divide", "size", "sum"]

        for name, condition in select.items():
            if isinstance(condition, str):
                select_query["project"]["fields"].append(
                    {"name": name, "key": condition}
                )
            elif isinstance(condition, dict):
                operator = condition.get("operator")
                key = condition.get("key")
                fields = condition.get("fields")

                if operator not in supported_operator:
                    raise ERROR_INVALID_PARAMETER(
                        key="query.select.operator",
                        reason=f"supported_operator = {supported_operator}",
                    )

                if key:
                    select_query["project"]["fields"].append(
                        {"name": name, "operator": operator, "key": key}
                    )
                elif fields:
                    select_query["project"]["fields"].append(
                        {"name": name, "operator": operator, "fields": fields}
                    )
                else:
                    raise ERROR_REQUIRED_PARAMETER(key="query.select.fields")

        return [select_query]

    @classmethod
    def _make_sort_query(cls, sort, group_fields, has_field_group):
        sort_query = {"sort": []}
        if has_field_group:
            group_field_names = []
            for group_field in group_fields:
                group_field_names.append(group_field["name"])

            for condition in sort:
                key = condition.get("key")
                desc = condition.get("desc", False)

                if key in group_field_names:
                    key = f"_total_{key}"

                sort_query["sort"].append({"key": key, "desc": desc})
        else:
            sort_query["sort"] = sort

        return [sort_query]

    @classmethod
    def _make_page_query(cls, page):
        page_query = []
        start = page.get("start")
        limit = page.get("limit")

        if limit:
            if start:
                page_query.append({"skip": start - 1})

            page_query.append({"limit": limit + 1})

        return page_query

    @classmethod
    def _make_date_filter(cls, date_field, key, operator):
        return [{"k": date_field, "v": key, "o": operator}]

    @classmethod
    def _parse_start_and_end_time(cls, key, value) -> date:
        if isinstance(value, date):
            return value
        elif isinstance(value, datetime):
            return value.date()
        elif isinstance(value, str):
            if len(value) == 4:
                date_format = "%Y"
                date_type = "YYYY"
            elif len(value) == 7:
                date_format = "%Y-%m"
                date_type = "YYYY-MM"
            else:
                date_format = "%Y-%m-%d"
                date_type = "YYYY-MM-DD"

            try:
                dt = datetime.strptime(value, date_format).date()

                if date_type == "YYYY":
                    if key == "start":
                        return datetime(dt.year, 1, 1)
                    else:
                        return datetime(dt.year, 1, 1) + relativedelta(years=1)
                elif date_type == "YYYY-MM":
                    if key == "start":
                        return datetime(dt.year, dt.month, 1)
                    else:
                        return datetime(dt.year, dt.month, 1) + relativedelta(months=1)
                else:
                    if key == "start":
                        return datetime(dt.year, dt.month, dt.day)
                    else:
                        return datetime(dt.year, dt.month, dt.day) + relativedelta(
                            days=1
                        )

            except Exception as e:
                raise ERROR_INVALID_PARAMETER_TYPE(key=key, type=date_type)
        else:
            raise ERROR_INVALID_PARAMETER(
                key=key, reason=f"{key} option should be datetime or str"
            )

    @classmethod
    def _convert_date_value(cls, date_value, date_field_format):
        if isinstance(date_field_format, str):
            return date_value.strftime(date_field_format)
        else:
            return date_value

    @classmethod
    def analyze(
        cls,
        *args,
        granularity=None,
        fields=None,
        select=None,
        group_by=None,
        field_group=None,
        filter=None,
        filter_or=None,
        page=None,
        sort=None,
        start=None,
        end=None,
        unwind=None,
        date_field="date",
        date_field_format="%Y-%m-%d",
        reference_filter=None,
        target="SECONDARY_PREFERRED",
        hint=None,
        allow_disk_use=False,
        return_type="dict",
        **kwargs,
    ):
        if fields is None:
            raise ERROR_REQUIRED_PARAMETER(key="fields")

        filter = filter or []
        filter_or = filter_or or []
        group_by = group_by or []
        sort = sort or []
        page = page or {}
        field_group = field_group or []
        cls._check_field_group(field_group)
        has_field_group = len(field_group) > 0
        page_limit = page.get("limit")

        if start:
            start_time: date = cls._parse_start_and_end_time("start", start)
            start_value = cls._convert_date_value(start_time, date_field_format)
            filter += cls._make_date_filter(date_field, start_value, "gte")

        if end:
            end_time: date = cls._parse_start_and_end_time("end", end)
            end_value = cls._convert_date_value(end_time, date_field_format)
            filter += cls._make_date_filter(date_field, end_value, "lt")

        group_keys = cls._make_group_keys(group_by, date_field, granularity)
        group_fields = cls._make_group_fields(fields)

        aggregate = []

        if unwind:
            aggregate.append({"unwind": unwind})

        aggregate.append({"group": {"keys": group_keys, "fields": group_fields}})

        query = {
            "filter": filter,
            "filter_or": filter_or,
            "aggregate": aggregate,
            "target": target,
            "hint": hint,
            "allow_disk_use": allow_disk_use,
            "return_type": return_type,
            "reference_filter": reference_filter,
        }

        if select:
            query["aggregate"] += cls._make_select_query(select)

        if has_field_group:
            query["aggregate"] += cls._make_field_group_query(
                group_keys, group_fields, field_group
            )

        if len(sort) > 0:
            query["aggregate"] += cls._make_sort_query(
                sort, group_fields, has_field_group
            )

        if page:
            query["aggregate"] += cls._make_page_query(page)

        response = cls.stat(**query)

        if return_type == "cursor":
            return response
        else:
            if page_limit:
                response["more"] = len(response["results"]) > page_limit
                response["results"] = response["results"][:page_limit]

            return response



PLUGIN_SOURCES = {
    'auth': 'spaceone.identity.plugin.auth.skeleton',
    'asset': 'spaceone.inventory.plugin.collector.skeleton',
    'metric': 'spaceone.monitoring.plugin.metric.skeleton',
    'log': 'spaceone.monitoring.plugin.log.skeleton',
    'webhook': 'spaceone.monitoring.plugin.webhook.skeleton',
    'cost': 'spaceone.cost_analysis.plugin.data_source.skeleton',
    'notification': 'spaceone.notification.plugin.protocol.skeleton',
}



# -*- coding: utf-8 -*-
import os
import sys
import pprint
import json
from xmlrunner.result import _XMLTestResult
from xmlrunner.result import _TestInfo

try:
    from google.protobuf.json_format import MessageToDict
except Exception as e:
    print(f'Failed to import: {str(e)}')

_PP = pprint.PrettyPrinter(indent=4)


def print_data(data, description=None):
    if os.environ.get('TEST_DEBUG', 'false') == 'true':
        print()
        if description:
            print(f'[ {description} ]')

        _PP.pprint(data)


def print_json(json_data, description=None):
    data = json.loads(json_data)
    print_data(data, description)


def print_message(message, description=None):
    data = MessageToDict(message, preserving_proto_field_name=True)
    print_data(data, description)


def testcase_name(test_method):
    fullpath = (test_method.id()).split('.')
    return fullpath[-1]


class RichTestInfo(_TestInfo):
    # Possible test outcomes
    (SUCCESS, FAILURE, ERROR, SKIP) = range(4)

    def __init__(self, test_result, test_method, outcome=SUCCESS, err=None, subTest=None, filename=None, lineno=None):
        super(RichTestInfo, self).__init__(test_result, test_method, outcome, err, subTest, filename, lineno)
        self.short_id = testcase_name(test_method)

    def result(self):
        if self.outcome == 0:
            result = "SUCCESS"
        else:
            result = (RichTestInfo.OUTCOME_ELEMENTS[self.outcome]).upper()
        return '{:<40s}{:<30s}{:>20s}{:>20f}'.format(self.short_id, self.test_name, result, self.elapsed_time)


class RichTestResult(_XMLTestResult):
    """
    A test result class that can express test results in a XML report.

    Used by XMLTestRunner.
    """

    def __init__(self, stream=sys.stderr, descriptions=1, verbosity=1,
                 elapsed_times=True, properties=None, infoclass=RichTestInfo):
        super(RichTestResult, self).__init__(stream, descriptions, verbosity, elapsed_times, properties, infoclass)

    def _prepare_callback(self, test_info, target_list, verbose_str,
                          short_str):
        """
            Appends a `infoclass` to the given target list and sets a callback
            method to be called by stopTest method.
            """
        test_info.filename = self.filename
        test_info.lineno = self.lineno
        target_list.append(test_info)

        def callback():
            """Prints the test method outcome to the stream, as well as
                the elapsed time.
                """

            test_info.test_finished()

            # Ignore the elapsed times for a more reliable unit testing
            if not self.elapsed_times:
                self.start_time = self.stop_time = 0

            self.stream.writeln(test_info.result())
            self.stream.flush()

        self.callback = callback



import json
import logging
import copy
from typing import Tuple, List

from spaceone.core import cache, config
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.core.auth.jwt import JWTAuthenticator, JWTUtil
from spaceone.core.transaction import get_transaction
from spaceone.core.handler import BaseAuthenticationHandler
from spaceone.core.error import ERROR_AUTHENTICATE_FAILURE

_LOGGER = logging.getLogger(__name__)


class SpaceONEAuthenticationHandler(BaseAuthenticationHandler):
    def __init__(self, handler_config):
        super().__init__(handler_config)
        self.identity_client = None
        self._initialize()

    def _initialize(self) -> None:
        self.identity_conn: SpaceConnector = SpaceConnector(service="identity")

    def verify(self, params: dict) -> None:
        token = self._get_token()
        domain_id = self._extract_domain_id(token)

        token_info = self._authenticate(token, domain_id)

        if token_info.get("typ") == "SYSTEM_TOKEN":
            self._update_system_meta(token_info)
        else:
            version = token_info.get("ver")
            if version not in ["2.0"]:
                raise ERROR_AUTHENTICATE_FAILURE(message="invalid token version.")

            owner_type = token_info.get("own")
            if owner_type == "APP":
                client_id = token_info.get("jti")
                domain_id = token_info.get("did")
                permissions, projects = self._check_app(client_id, domain_id)
                token_info["permissions"] = copy.deepcopy(permissions)
                token_info["projects"] = copy.deepcopy(projects)

            self._update_meta(token_info)

    @cache.cacheable(key="handler:authentication:{domain_id}:public-key", alias="local")
    def _get_public_key(self, domain_id: str) -> str:
        system_token = config.get_global("TOKEN")

        _LOGGER.debug(f"[_get_public_key] get jwk from identity service: {domain_id}")
        response = self.identity_conn.dispatch(
            "Domain.get_public_key", {"domain_id": domain_id}, token=system_token
        )

        return response["public_key"]

    @cache.cacheable(
        key="handler:authentication:{domain_id}:client:{client_id}", alias="local"
    )
    def _check_app(self, client_id: str, domain_id: str) -> Tuple[List[str], List[str]]:
        system_token = config.get_global("TOKEN")

        _LOGGER.debug(f"[_check_app] check app from identity service: {client_id}")
        response = self.identity_conn.dispatch(
            "App.check",
            {
                "client_id": client_id,
                "domain_id": domain_id,
            },
            token=system_token,
        )

        return response.get("permissions", []), response.get("projects", [])

    def _authenticate(self, token: str, domain_id: str) -> dict:
        public_key = self._get_public_key(domain_id)
        return JWTAuthenticator(json.loads(public_key)).validate(token)

    @staticmethod
    def _get_token() -> str:
        transaction = get_transaction()
        token = transaction.meta.get("token")
        if not isinstance(token, str) or len(token) == 0:
            raise ERROR_AUTHENTICATE_FAILURE(message="empty token provided.")

        return token

    @staticmethod
    def _extract_domain_id(token):
        try:
            decoded = JWTUtil.unverified_decode(token)
        except Exception:
            _LOGGER.debug(f"[_extract_domain_id] failed to decode token: {token[:10]}")
            raise ERROR_AUTHENTICATE_FAILURE(message="failed to decode token.")

        if domain_id := decoded.get("did"):
            return domain_id
        else:
            raise ERROR_AUTHENTICATE_FAILURE(message="empty domain_id provided.")

    def _update_meta(self, token_info: dict) -> None:
        """
        Args:
            token_info(dict): {
                'iss': 'str',   # issuer (spaceone.identity)
                'rol': 'str',   # role type (SYSTEM_TOKEN | DOMAIN_ADMIN | WORKSPACE_OWNER | WORKSPACE_MEMBER | USER )
                'typ': 'str',   # token type (ACCESS_TOKEN | REFRESH_TOKEN | CLIENT_SECRET)
                'own': 'str',   # owner (USER | APP)
                'did': 'str',   # domain_id
                'wid': 'str',   # workspace_id, Optional
                'aud': 'str',   # audience (user_id | app_id)
                'exp': 'int',   # expiration time
                'iat': 'int',   # issued at
                'jti': 'str',   # jwt id (token_key | client_id), Optional
                'permissions': 'list',  # permissions, Optional
                'projects': 'list',     # project_ids, if workspace member, Optional
                'user_groups': 'list',  # user_group_ids, if workspace owner or member, Optional
                'injected_params': 'dict',  # injected parameters, override parameters, Optional
                'ver': 'str',   # jwt version
        """

        token_type = token_info.get("typ")
        role_type = token_info.get("rol")
        owner_type = token_info.get("own")
        audience = token_info.get("aud")
        domain_id = token_info.get("did")
        workspace_id = token_info.get("wid")
        permissions = token_info.get("permissions")
        projects = token_info.get("projects")
        user_groups = token_info.get("user_groups")
        injected_params = token_info.get("injected_params")

        self.transaction.set_meta("authorization.token_type", token_type)
        self.transaction.set_meta("authorization.role_type", role_type)
        self.transaction.set_meta("authorization.owner_type", owner_type)
        self.transaction.set_meta("authorization.domain_id", domain_id)
        self.transaction.set_meta("authorization.audience", audience)
        self.transaction.set_meta("authorization.workspace_id", workspace_id)
        self.transaction.set_meta("authorization.permissions", permissions)
        self.transaction.set_meta("authorization.projects", projects)
        self.transaction.set_meta("authorization.user_groups", user_groups)
        self.transaction.set_meta("authorization.injected_params", injected_params)

        if owner_type == "USER":
            self.transaction.set_meta("authorization.user_id", audience)
        elif owner_type == "APP":
            self.transaction.set_meta("authorization.app_id", audience)

    def _update_system_meta(self, token_info: dict) -> None:
        """
        Args:
            token_info(dict): {
                'iss': 'str',   # issuer (spaceone.identity)
                'typ': 'str',   # token type (SYSTEM_TOKEN)
                'own': 'str',   # owner (SYSTEM)
                'did': 'str',   # domain_id (domain-root)
                'aud': 'str',   # audience (root_domain_user_id)
                'iat': 'int',   # issued at
                'jti': 'str',   # jwt id
                'ver': 'str',   # jwt version
        """

        token_type = token_info.get("typ")
        owner_type = token_info.get("own")
        audience = token_info.get("aud")
        domain_id = self.transaction.get_meta("x_domain_id")
        workspace_id = self.transaction.get_meta("x_workspace_id")

        self.transaction.set_meta("authorization.token_type", token_type)
        self.transaction.set_meta("authorization.role_type", "SYSTEM_TOKEN")
        self.transaction.set_meta("authorization.owner_type", owner_type)
        self.transaction.set_meta("authorization.domain_id", domain_id)
        self.transaction.set_meta("authorization.audience", audience)
        self.transaction.set_meta("authorization.workspace_id", workspace_id)



# -*- coding: utf-8 -*-


from spaceone.core.scheduler.server import serve
from spaceone.core.scheduler.scheduler import *

__all__ = ['serve', 'IntervalScheduler', 'CronScheduler', 'HourlyScheduler']



import logging
import logging.config
import copy

from spaceone.core import config
from spaceone.core.error import *
from spaceone.core.utils import *
from spaceone.core.logger.filters import *

__all__ = ["set_logger"]

DEFAULT_LOGGER = "spaceone"

HANDLER_DEFAULT_CONSOLE = {
    "class": "logging.StreamHandler",
    "formatter": "standard",
    "filters": ["transaction", "masking", "exclude", "parameter", "traceback"],
}

HANDLER_DEFAULT_FILE = {
    "class": "logging.handlers.RotatingFileHandler",
    "filename": "",
    "filters": [
        "transaction",
        "masking",
        "exclude",
        "error_message",
        "parameter_log",
        "message",
        "traceback_log",
    ],
    "formatter": "file",
    "maxBytes": 10485760,  # 10 MB
    "backupCount": 10,
}

HANDLER_DEFAULT_TMPL = {
    "console": HANDLER_DEFAULT_CONSOLE,
    "file": HANDLER_DEFAULT_FILE,
}

FORMATTER_DEFAULT_TMPL = {
    "standard": {
        "format": "%(asctime)s.%(msecs)03dZ [%(levelname)s] %(trace_id)s %(domain_id)s %(audience)s %(role_type)s %(workspace_id)s %(tnx_method)s (%(filename)s:%(lineno)d) %(message)s %(params)s %(traceback)s",
        "datefmt": "%Y-%m-%dT%H:%M:%S",
    },
    "file": {
        "format": '{"time": "%(asctime)s.%(msecs)03dZ", "level": "%(levelname)s", "peer": "%(peer)s", "trace_id": "%(trace_id)s", "domain_id": "%(domain_id)s", "audience": "%(audience)s", "role_type": "%(role_type)s", "tnx_method": "%(tnx_method)s", "file_name": "%(filename)s", "line": %(lineno)d, "parameter": %(params_log)s, "message": %(msg_dump)s, "error": { "code": "%(error_code)s", "message": "%(error_message)s", "traceback": %(traceback_log)s }}',
        "datefmt": "%Y-%m-%dT%H:%M:%S",
    },
}

FILTER_DEFAULT_TMPL = {
    "masking": {"()": MaskingFilter, "rules": {}},
    "transaction": {"()": TransactionFilter},
    "traceback": {"()": TracebackFilter},
    "traceback_log": {"()": TracebackLogFilter},
    "parameter": {"()": ParameterFilter},
    "parameter_log": {"()": ParameterLogFilter},
    "error_message": {"()": ErrorFilter},
    "message": {"()": MessageJsonFilter},
    "exclude": {
        "()": ExcludeFilter,
        "rules": {
            "tnx_method": [],
        },
    },
}

_LOGGER = {
    "version": 1,
    "formatters": {},
    "filters": {},
    "handlers": {"console": HANDLER_DEFAULT_CONSOLE},
    "loggers": {},
}

LOGGER_DEFAULT_TMPL = {"level": "DEBUG", "propagate": True, "handlers": ["console"]}


def set_logger(transaction=None):
    _set_config(transaction)
    logging.config.dictConfig(_LOGGER)


def _set_default_logger(default_logger):
    _LOGGER["loggers"] = {default_logger: LOGGER_DEFAULT_TMPL}
    _LOGGER["formatters"] = FORMATTER_DEFAULT_TMPL


def _set_loggers(loggers):
    for _logger in loggers:
        _LOGGER["loggers"][_logger] = deep_merge(
            loggers[_logger], copy.deepcopy(LOGGER_DEFAULT_TMPL)
        )


def _set_transaction_filter(transaction):
    if transaction:
        _LOGGER["filters"]["transaction"]["transaction"] = transaction


def _set_handlers(handlers):
    for _handler in handlers:
        _default = copy.deepcopy(HANDLER_DEFAULT_TMPL)

        if "type" in handlers[_handler]:
            if handlers[_handler]["type"] not in HANDLER_DEFAULT_TMPL:
                raise ERROR_LOG_CONFIG(reason="Logger handler type is not supported")

            _default = copy.deepcopy(HANDLER_DEFAULT_TMPL[handlers[_handler]["type"]])

        _default = deep_merge(handlers[_handler], _default)

        if "type" in _default:
            del _default["type"]

        _LOGGER["handlers"][_handler] = _default


def _set_formatters(formatters):
    for _formatter in formatters:
        _default = {}

        if "type" in formatters[_formatter]:
            if formatters[_formatter]["type"] not in FORMATTER_DEFAULT_TMPL:
                raise ERROR_LOG_CONFIG(reason="Logger formatter type is not supported")

            _default = copy.deepcopy(
                FORMATTER_DEFAULT_TMPL[formatters[_formatter]["type"]]
            )

        _default = deep_merge(formatters[_formatter]["args"], _default)

        if "type" in _default:
            del _default["type"]

        _LOGGER["formatters"][_formatter] = _default


def _set_filters(filters):
    _LOGGER["filters"] = deep_merge(filters, copy.deepcopy(FILTER_DEFAULT_TMPL))


def _set_config(transaction):
    global_log_conf = config.get_global("LOG", {})

    _set_default_logger(DEFAULT_LOGGER)

    if "loggers" in global_log_conf:
        _set_loggers(global_log_conf["loggers"])

    if "handlers" in global_log_conf:
        _set_handlers(global_log_conf["handlers"])

    if "formatters" in global_log_conf:
        _set_formatters(global_log_conf["formatters"])

    _set_filters(global_log_conf.get("filters", {}))
    _set_transaction_filter(transaction)



import logging
from spaceone.core import config
from spaceone.core.error import ERROR_BASE

_LOGGER = logging.getLogger(__name__)

DEFAULT_POOL = 8


class Server(object):
    def __init__(self, service, config):
        self.service = service
        self.config = config
        self.queues = {}
        self.schedulers = {}
        self.workers = {}

    def start(self):
        ###################
        # Queues
        ###################
        if 'QUEUES' not in self.config:
            _LOGGER.error('QUEUES is not configured')
            # TODO: exit
        # Load Queue config
        self.queues = self.config['QUEUES']
        _LOGGER.info(f'[Server] queue: {self.queues}')

        ####################
        # Start Schedulers
        ####################
        if 'SCHEDULERS' not in self.config:
            _LOGGER.warn('SCHEDULERS is not configured')
            _LOGGER.warn('Scheduling will be triggered remotely')
            self.config['SCHEDULERS'] = {}

        for (name, conf) in self.config['SCHEDULERS'].items():
            _LOGGER.info(f'[Server] name: {name}')
            _LOGGER.info(f'[Server] config: {conf}')
            params = conf.copy()
            if 'backend' not in conf:
                _LOGGER.debug('Backend is not specified')
                # TODO: ERROR
            backend = conf['backend']
            del params['backend']
            if 'queue' not in conf:
                _LOGGER.debug('Queue is not specified')
                # TODO: ERROR
            self.schedulers[name] = self._create_process(backend, params)

        #######################
        # Start Workers
        #######################
        if 'WORKERS' not in self.config:
            _LOGGER.warn('WORKER is not configured')
            _LOGGER.warn('May be schduler only')
            self.config['WORKERS'] = {}
            # TODO: exit
        # Load Worker config
        for (name, conf) in self.config['WORKERS'].items():
            params = conf.copy()
            if 'backend' not in conf:
                _LOGGER.debug('Backend is not specified')
                # TODO: ERROR
            backend = conf['backend']
            del params['backend']
            if 'queue' not in conf:
                _LOGGER.debug('Queue is not specified')
                # TODO: ERROR
            # Support Worker pool (default: 8)
            # self.queue is instance of Queue
            pool = params.get('pool', DEFAULT_POOL)
            _LOGGER.debug(f'[start] create thread pool: {pool}')
            for index in range(pool):
                worker_name = f'{name}_{index}'
                _LOGGER.debug(f'[start] create {worker_name}')
                self.workers[worker_name] = self._create_process(backend, params)

        # Start All threads
        # start worker
        for (k, v) in self.workers.items():
            v.start()

        # start scheduler
        for (k, v) in self.schedulers.items():
            v.start()

    def _create_process(self, backend, params):
        # create scheduler
        _LOGGER.debug(params)
        try:
            module_name, class_name = backend.rsplit('.', 1)
            scheduler_module = __import__(module_name, fromlist=[class_name])
            return getattr(scheduler_module, class_name)(**params)
        except ERROR_BASE as error:
            _LOGGER.error({error})
            raise error
        except Exception as e:
            _LOGGER.error({e})
            raise e


def serve():
    # Load scheduler config
    # Create Scheduler threads
    # start Scheduler
    conf = config.get_global()

    server = Server(config.get_service(), conf)
    server.start()



import logging

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from spaceone.core import config

_LOGGER = logging.getLogger(__name__)


def _get_router_conf():
    package = config.get_package()
    router_conf_module = __import__(
        f"{package}.conf.router_conf", fromlist=["router_conf"]
    )
    return getattr(router_conf_module, "ROUTER", [])


def _get_sub_app_conf():
    package = config.get_package()
    router_conf_module = __import__(
        f"{package}.conf.router_conf", fromlist=["router_conf"]
    )
    return getattr(router_conf_module, "SUB_APP", {})


def _get_router(path):
    try:
        module_path, router_name = path.split(":")
        module_name = module_path.rsplit(".")[-1:]
        router_module = __import__(module_path, fromlist=[module_name])
        return getattr(router_module, router_name)
    except Exception as e:
        _LOGGER.warning(
            f"[_get_router] Invalid router path. (router_path = {path})", exc_info=True
        )


def _mount_sub_apps(app, sub_apps):
    sub_app_conf = _get_sub_app_conf()
    for sub_app_name, sub_app_options in sub_app_conf.items():
        sub_app = sub_apps.get(sub_app_name)
        if sub_app:
            sub_app_path = sub_app_options.get("path")
            app.mount(path=sub_app_path, app=sub_app)


def _create_sub_app(sub_app_options):
    title = sub_app_options.get("title", "FastAPI")
    description = sub_app_options.get("description", "")
    contact = sub_app_options.get("contact", {})

    return FastAPI(title=title, description=description, contact=contact)


def _include_routers(app):
    # Include All Routers from router_conf.py
    routers_conf = _get_router_conf()
    sub_apps_conf = _get_sub_app_conf()

    sub_apps = {}
    all_routers_path = []

    # App Routers
    for router_conf in routers_conf:
        sub_app_name = router_conf.get("sub_app")
        router_path = router_conf.get("router_path")
        router_options = router_conf.get("router_options", {})

        if router_path is None:
            _LOGGER.warning(
                f"[include_routers] Undefined router_path. (router = {router_conf})"
            )
            continue

        if sub_app_name in sub_apps_conf:
            if sub_app_name not in sub_apps:
                sub_app_options = sub_apps_conf[sub_app_name]
                sub_apps[sub_app_name] = _create_sub_app(sub_app_options)

            _append_router(sub_apps[sub_app_name], router_path, router_options)
        else:
            _append_router(app, router_path, router_options)

        all_routers_path.append(router_path)

    # Extension Routers
    ext_routers = config.get_global("REST_EXTENSION_ROUTERS", [])
    for router in ext_routers:
        router_path = router.get("router_path")
        router_options = router.get("router_options", {})
        _append_router(app, router_path, router_options)

        all_routers_path.append(router_path)

    # Mount Sub Applications
    _mount_sub_apps(app, sub_apps)

    all_routers_path_str = "\n\t - ".join(all_routers_path)
    _LOGGER.debug(f"Loaded Routers: \n\t - {all_routers_path_str}")

    return app


def _append_router(app, router_path, router_options):
    router = _get_router(router_path)

    app.include_router(router, **router_options)


def _add_middlewares(app):
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    return app


def _init_fast_api():
    global_conf = config.get_global()

    return FastAPI(
        title=global_conf.get("REST_TITLE", "Document"),
        version="x.y.z",
        # version=server_info.get_version(),
        contact=global_conf.get("REST_CONTACT", {}),
        description=global_conf.get("REST_DESCRIPTION", ""),
    )


def fast_api_app():
    app = _init_fast_api()
    app = _add_middlewares(app)
    app = _include_routers(app)
    return app


def serve(app_path: str = None):
    conf = config.get_global()
    app_path = conf["REST_APP_PATH"]

    uvicorn_options = conf.get("UVICORN_OPTIONS", {})

    _LOGGER.info(
        f"Start REST Server ({config.get_service()}): "
        f'host={conf["HOST"]} port={conf["PORT"]} options={uvicorn_options}'
    )

    uvicorn.run(
        "spaceone.core.fastapi.server:fast_api_app",
        host=conf["HOST"],
        port=conf["PORT"],
        **uvicorn_options,
    )



import logging
import json


class ErrorFilter(logging.Filter):
    def filter(self, record):
        if getattr(record, "error_code", None) is None:
            record.error_code = ""

        if getattr(record, "error_message", None) is None:
            record.error_message = ""

        if getattr(record, "traceback", None) is None:
            record.traceback = ""
        else:
            record.traceback = json.dumps(record.traceback)

        return True



import threading
import traceback
import logging
from threading import local

from spaceone.core import utils
from opentelemetry import trace
from opentelemetry.trace import format_trace_id
from opentelemetry.trace.span import TraceFlags

__all__ = [
    "LOCAL_STORAGE",
    "get_transaction",
    "create_transaction",
    "delete_transaction",
    "Transaction",
]

_LOGGER = logging.getLogger(__name__)
LOCAL_STORAGE = local()


class Transaction(object):
    def __init__(
        self,
        service: str = None,
        resource: str = None,
        verb: str = None,
        trace_id: str = None,
        meta=None,
    ):
        self._id = None
        self._thread_id = str(threading.current_thread().ident)
        self._service = service
        self._resource = resource
        self._verb = verb
        self._rollbacks = []
        self._init_meta(meta)
        self._set_trace_id(trace_id)

    def __repr__(self) -> str:
        return f"<Transaction ({self._resource}.{self._verb})>"

    def _set_trace_id(self, trace_id: str = None) -> None:
        if trace_id:
            self._id = trace_id
        else:
            self._id = format_trace_id(utils.generate_trace_id())

    def _init_meta(self, meta: dict = None):
        if meta:
            self._meta = meta.copy()
        else:
            self._meta = {}

    @property
    def id(self) -> str:
        return self._id

    @property
    def thread_id(self) -> str:
        return self._thread_id

    @property
    def service(self) -> str:
        return self._service

    @property
    def resource(self) -> str:
        return self._resource

    @property
    def verb(self) -> str:
        return self._verb

    def add_rollback(self, fn: callable, *args, **kwargs) -> None:
        self._rollbacks.insert(0, {"fn": fn, "args": args, "kwargs": kwargs})

    def execute_rollback(self) -> None:
        for rollback in self._rollbacks:
            try:
                rollback["fn"](*rollback["args"], **rollback["kwargs"])
            except Exception as e:
                _LOGGER.info(f"[ROLLBACK-ERROR] {self}: {e}")
                _LOGGER.info(traceback.format_exc())

    @property
    def meta(self) -> dict:
        return self._meta

    def set_meta(self, key: str, value: any) -> None:
        self._meta[key] = value

    def get_meta(self, key: str, default: any = None):
        return self._meta.get(key, default)


def get_transaction(is_create: bool = True) -> [Transaction, None]:
    current_span_context = trace.get_current_span().get_span_context()
    thread_id = str(threading.current_thread().ident)

    if current_span_context.trace_flags == TraceFlags.SAMPLED:
        trace_id_from_current_span = format_trace_id(current_span_context.trace_id)
        return getattr(LOCAL_STORAGE, trace_id_from_current_span, None)
    elif hasattr(LOCAL_STORAGE, thread_id):
        return getattr(LOCAL_STORAGE, thread_id, None)
    elif is_create:
        return create_transaction(thread_id=thread_id)
    else:
        return None


def create_transaction(
    service: str = None,
    resource: str = None,
    verb: str = None,
    trace_id: str = None,
    meta: dict = None,
    thread_id: str = None,
) -> Transaction:
    transaction = Transaction(service, resource, verb, trace_id, meta)

    if thread_id:
        setattr(LOCAL_STORAGE, thread_id, transaction)
    else:
        setattr(LOCAL_STORAGE, transaction.id, transaction)

    return transaction


def delete_transaction() -> None:
    if transaction := get_transaction(is_create=False):
        if hasattr(LOCAL_STORAGE, transaction.id):
            delattr(LOCAL_STORAGE, transaction.id)

    thread_id = str(threading.current_thread().ident)
    if hasattr(LOCAL_STORAGE, thread_id):
        delattr(LOCAL_STORAGE, thread_id)



import re
import functools
import types
from typing import Union
from dateutil.parser import parse
from datetime import datetime, date
from typing import get_type_hints
from pydantic import ValidationError, BaseModel

from spaceone.core import utils
from spaceone.core.error import *

__all__ = [
    "convert_model",
    "change_value_by_rule",
    "change_only_key",
    "check_required",
    "set_query_page_limit",
    "append_query_filter",
    "change_tag_filter",
    "append_keyword_filter",
    "change_timestamp_value",
    "change_date_value",
    "change_timestamp_filter",
    "check_query_filter",
]


def _raise_pydantic_error(e: ValidationError):
    for error in e.errors():
        if error["type"] == "value_error.missing":
            raise ERROR_REQUIRED_PARAMETER(key=", ".join(error["loc"]))
        elif error["type"].startswith("type_error"):
            raise ERROR_INVALID_PARAMETER_TYPE(
                key=", ".join(error["loc"]), type=error["type"].split(".")[1]
            )
        else:
            raise ERROR_INVALID_PARAMETER(
                key=", ".join(error["loc"]), reason=error["msg"]
            )


def _generate_response(response_iterator: types.GeneratorType) -> types.GeneratorType:
    for response in response_iterator:
        if isinstance(response, BaseModel):
            response = response.dict()

        yield response


def convert_model(func) -> callable:
    @functools.wraps(func)
    def wrapped_func(self, params: dict) -> Union[dict, types.GeneratorType]:
        type_hints = get_type_hints(func)
        params_hint = type_hints.get("params")

        if params_hint and isinstance(params, dict):
            try:
                params = params_hint(**params)
            except ValidationError as e:
                _raise_pydantic_error(e)

        response_or_iterator = func(self, params)

        if isinstance(response_or_iterator, types.GeneratorType):
            return _generate_response(response_or_iterator)
        else:
            if isinstance(response_or_iterator, BaseModel):
                response_or_iterator = response_or_iterator.dict()

        return response_or_iterator

    return wrapped_func


def change_value_by_rule(rule: str, param_key: str, value: any = None) -> any:
    def wrapper(func: callable) -> callable:
        @functools.wraps(func)
        def wrapped_func(cls, params: dict) -> Union[dict, types.GeneratorType]:
            if param_key in params:
                param_value = params[param_key]
                if rule == "APPEND":
                    if isinstance(param_value, list):
                        param_value.append(value)
                    else:
                        param_value = [param_value, value]

                params[param_key] = param_value

            return func(cls, params)

        return wrapped_func

    return wrapper


def change_only_key(change_rule, key_path="only") -> callable:
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            only_keys = change_rule.keys()
            only = utils.get_dict_value(params, key_path, [])
            new_only = []
            for key in only:
                key_match = key.split(".", 1)[0]
                if key_match in only_keys:
                    if change_rule[key_match] not in new_only:
                        new_only.append(change_rule[key_match])
                else:
                    new_only.append(key)

            utils.change_dict_value(params, key_path, new_only)
            return func(cls, params)

        return wrapped_func

    return wrapper


def check_required(required_keys) -> callable:
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            for key in required_keys:
                value = utils.get_dict_value(params, key)
                if value is None:
                    raise ERROR_REQUIRED_PARAMETER(key=key)

            return func(cls, params)

        return wrapped_func

    return wrapper


def set_query_page_limit(default_limit) -> callable:
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            query = params.get("query", {})
            query["page"] = query.get("page", {})
            page_limit = query["page"].get("limit", 1000)

            if page_limit > default_limit:
                page_limit = default_limit

            query["page"]["limit"] = page_limit
            params["query"] = query

            return func(cls, params)

        return wrapped_func

    return wrapper


def append_query_filter(filter_keys) -> callable:
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            query = params.get("query", {})
            query["filter"] = query.get("filter", [])

            for key in filter_keys:
                if key in params:
                    if isinstance(params[key], list):
                        query["filter"].append({"k": key, "v": params[key], "o": "in"})
                    else:
                        query["filter"].append({"k": key, "v": params[key], "o": "eq"})

            params["query"] = query

            return func(cls, params)

        return wrapped_func

    return wrapper


def change_tag_filter(tag_key="tags") -> callable:
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            params["query"] = params.get("query", {})
            change_filter = []
            for condition in params["query"].get("filter", []):
                key = condition.get("key", condition.get("k"))
                if key and key.startswith(f"{tag_key}."):
                    value = condition.get("value", condition.get("v"))
                    operator = condition.get("operator", condition.get("o"))
                    tag_key_split = key.split(".", 1)
                    change_filter.append(
                        {
                            "key": "tags",
                            "value": {
                                "key": tag_key_split[1],
                                "value": _change_match_query(
                                    operator, value, condition
                                ),
                            },
                            "operator": "match",
                        }
                    )
                else:
                    change_filter.append(condition)

            if "only" in params["query"]:
                change_only = []
                for key in params["query"].get("only", []):
                    if key.startswith("tags.") and "tags" not in change_only:
                        change_only.append("tags")
                    else:
                        change_only.append(key)

                params["query"]["only"] = change_only

            params["query"]["filter"] = change_filter
            return func(cls, params)

        return wrapped_func

    return wrapper


def _change_match_query(operator, value, condition) -> callable:
    if operator == "eq":
        return value
    elif operator == "not":
        return {"$ne": value}
    elif operator == "in":
        if not isinstance(value, list):
            raise ERROR_OPERATOR_LIST_VALUE_TYPE(operator=operator, condition=condition)

        return {"$in": value}
    elif operator == "not_in":
        if not isinstance(value, list):
            raise ERROR_OPERATOR_LIST_VALUE_TYPE(operator=operator, condition=condition)

        return {"$nin": value}
    elif operator == "contain":
        return re.compile(value, re.IGNORECASE)
    elif operator == "not_contain":
        return {"$not": re.compile(value, re.IGNORECASE)}
    elif operator == "contain_in":
        if not isinstance(value, list):
            raise ERROR_OPERATOR_LIST_VALUE_TYPE(operator=operator, condition=condition)

        return {"$in": value}
    elif operator == "not_contain_in":
        if not isinstance(value, list):
            raise ERROR_OPERATOR_LIST_VALUE_TYPE(operator=operator, condition=condition)

        return {"$nin": value}
    else:
        raise ERROR_DB_QUERY(reason="Filter operator is not supported.")


def append_keyword_filter(keywords=None) -> callable:
    if keywords is None:
        keywords = []

    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            query = params.get("query", {})
            if "keyword" in query:
                query["filter_or"] = query.get("filter_or", [])

                keyword = query["keyword"].strip()
                if len(keyword) > 0:
                    for key in keywords:
                        query["filter_or"].append(
                            {
                                "k": key,
                                "v": list(filter(None, keyword.split(" "))),
                                "o": "contain_in",
                            }
                        )

                del query["keyword"]
                params["query"] = query

            return func(cls, params)

        return wrapped_func

    return wrapper


def change_timestamp_value(
        timestamp_keys=None, timestamp_format="google_timestamp"
) -> callable:
    if timestamp_keys is None:
        timestamp_keys = []

    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            change_params = {}

            for key, value in params.items():
                if key in timestamp_keys:
                    if not _is_null(value):
                        value = _convert_datetime_from_timestamp(
                            value, key, timestamp_format
                        )
                        change_params[key] = value
                else:
                    change_params[key] = value

            return func(cls, change_params)

        return wrapped_func

    return wrapper


def change_date_value(date_keys=None, date_format="%Y-%m-%d") -> callable:
    if date_keys is None:
        date_keys = []

    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            change_params = {}

            for key, value in params.items():
                if key in date_keys:
                    if not _is_null(value):
                        value = _convert_date_from_string(value, key, date_format)
                        change_params[key] = value
                else:
                    change_params[key] = value

            return func(cls, change_params)

        return wrapped_func

    return wrapper


def change_timestamp_filter(
        filter_keys=None, timestamp_format="google_timestamp"
) -> callable:
    if filter_keys is None:
        filter_keys = []

    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            query = params.get("query", {})
            query_filter = query.get("filter")
            query_filter_or = query.get("filter_or")
            if query_filter:
                query["filter"] = _change_timestamp_condition(
                    query_filter, filter_keys, "filter", timestamp_format
                )

            if query_filter_or:
                query["filter_or"] = _change_timestamp_condition(
                    query_filter_or, filter_keys, "filter_or", timestamp_format
                )

            params["query"] = query

            return func(cls, params)

        return wrapped_func

    return wrapper


def _is_null(value) -> bool:
    if value is None or str(value).strip() == "":
        return True

    return False


def _change_timestamp_condition(
        query_filter, filter_keys, filter_type, timestamp_format
) -> list:
    change_filter = []

    for condition in query_filter:
        key = condition.get("k") or condition.get("key")
        value = condition.get("v") or condition.get("value")
        operator = condition.get("o") or condition.get("operator")

        if key in filter_keys:
            value = _convert_datetime_from_timestamp(
                value, f"query.{filter_type}.{key}", timestamp_format
            )

        change_filter.append({"key": key, "value": value, "operator": operator})

    return change_filter


def _convert_datetime_from_timestamp(timestamp, key, timestamp_format) -> datetime:
    type_message = (
        "google.protobuf.Timestamp({seconds: <second>, nanos: <nano second>})"
    )

    try:
        if timestamp_format == "iso8601":
            type_message = "ISO 8601(YYYY-MM-DDThh:mm:ssTZD)"
            return parse(timestamp)
        else:
            seconds = timestamp["seconds"]
            nanos = timestamp.get("nanos")

            # TODO: change nanoseconds to timestamp

            return datetime.utcfromtimestamp((int(seconds)))
    except Exception as e:
        raise ERROR_INVALID_PARAMETER_TYPE(key=key, type=type_message)


def _convert_date_from_string(date_str, key, date_format) -> date:
    try:
        return datetime.strptime(date_str, date_format).date()
    except Exception as e:
        raise ERROR_INVALID_PARAMETER_TYPE(key=key, type=date_format)


def check_query_filter(keywords=None) -> callable:
    if keywords is None:
        keywords = []

    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(cls, params):
            query = params.get("query", {})
            if "filter" in query:
                for filters in query["filter"]:
                    key = filters.get("key", filters.get("k"))
                    if key in keywords:
                        raise ERROR_INVALID_PARAMETER(
                            key=key, reason="Include secure parameter"
                        )

            if "group_by" in query:
                for group_bys in query["group_by"]:
                    key = group_bys.get("key", group_bys.get("k"))
                    if key in keywords:
                        raise ERROR_INVALID_PARAMETER(
                            key=key, reason="Include secure parameter"
                        )

            if "fields" in query:
                value = query["fields"].get("value", query["fields"].get("v"))
                key = value.get("key", value.get("k"))
                if key in keywords:
                    raise ERROR_INVALID_PARAMETER(
                        key=key, reason="Include secure parameter"
                    )

            if "distinct" in query:
                key = query["distinct"]
                if key in keywords:
                    raise ERROR_INVALID_PARAMETER(
                        key=key, reason="Include secure parameter"
                    )

            return func(cls, params)

        return wrapped_func

    return wrapper



from spaceone.core.logger.filters.error import ErrorFilter
from spaceone.core.logger.filters.exclude import ExcludeFilter
from spaceone.core.logger.filters.masking import MaskingFilter
from spaceone.core.logger.filters.message import MessageJsonFilter
from spaceone.core.logger.filters.parameter import ParameterFilter, ParameterLogFilter
from spaceone.core.logger.filters.transaction import TransactionFilter
from spaceone.core.logger.filters.traceback import TracebackFilter, TracebackLogFilter


class ERROR_BASE(Exception):
    _status_code = "INTERNAL"
    _error_code = "ERROR_BASE"
    _message = "Base Error Class"
    _meta = {}

    def __init__(self, _error_code=None, _meta=None, **kwargs):
        if isinstance(_meta, dict):
            self._meta = _meta

        if _error_code:
            self._error_code = _error_code.strip()
        else:
            self._error_code = self.__class__.__name__

        try:
            self._message = self.message.format(**kwargs)
        except Exception:
            raise ERROR_MESSAGE_FORMAT(
                error_class=self.__class__.__name__, key=str(kwargs)
            )

    @property
    def message(self):
        return self._message

    @property
    def error_code(self):
        return self._error_code

    @error_code.setter
    def error_code(self, value):
        self._error_code = value

    @property
    def status_code(self):
        return self._status_code

    @status_code.setter
    def status_code(self, value):
        self._status_code = value

    @property
    def meta(self):
        return self._meta

    @meta.setter
    def meta(self, key, value):
        self._meta[key] = value

    def _repr(self):
        return (
            "\n"
            f"\terror_code = {self._error_code}\n"
            f"\tstatus_code = {self._status_code}\n"
            f"\tmessage = {self._message}"
        )

    def __repr__(self):
        return self._repr()

    def __str__(self):
        return self._repr()


class ERROR_INVALID_ARGUMENT(ERROR_BASE):
    _status_code = "INVALID_ARGUMENT"
    _message = "Argument is invalid."


class ERROR_REQUIRED_X_DOMAIN_ID(ERROR_INVALID_ARGUMENT):
    _message = "System token requires 'x_domain_id' in metadata."


class ERROR_REQUIRED_PARAMETER(ERROR_INVALID_ARGUMENT):
    _message = "Required parameter. (key = {key})"


class ERROR_INVALID_PARAMETER_TYPE(ERROR_INVALID_ARGUMENT):
    _message = "Parameter type is invalid. (key = {key}, type = {type})"


class ERROR_INVALID_PARAMETER(ERROR_INVALID_ARGUMENT):
    _message = "Parameter is invalid. (key = {key}, reason = {reason})"


class ERROR_NOT_FOUND(ERROR_INVALID_ARGUMENT):
    _status_code = "NOT_FOUND"
    _message = "Resource cannot be found or access is denied. ({key} = {value})"


class ERROR_NOT_UNIQUE(ERROR_INVALID_ARGUMENT):
    _status_code = "ALREADY_EXISTS"
    _message = "Tried to save duplicate unique key. ({key} = {value})"


class ERROR_SAVE_UNIQUE_VALUES(ERROR_NOT_UNIQUE):
    _message = "Tried to save duplicate unique values. (keys = {keys})"


class ERROR_EXIST_RESOURCE(ERROR_INVALID_ARGUMENT):
    _message = "'{child}' resources is existed in {parent}."


class ERROR_OPERATOR_VALUE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = (
        "The value of '{operator} operator' does not support list type. ({condition})"
    )


class ERROR_OPERATOR_LIST_VALUE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = "The value of '{operator} operator' must be a list type. ({condition})"


class ERROR_OPERATOR_DICT_VALUE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = (
        "The value of '{operator} operator' must be a dictionary type. ({condition})"
    )


class ERROR_OPERATOR_BOOLEAN_TYPE(ERROR_INVALID_ARGUMENT):
    _message = (
        "The value of '{operator} operator' must be a boolean type. ({condition})"
    )


class ERROR_INVALID_FORMAT(ERROR_INVALID_ARGUMENT):
    _message = "Value format is invalid. ({key} = {value} ! {rule})"


class ERROR_JSON_FORMAT(ERROR_INVALID_ARGUMENT):
    _message = "JSON format is invalid. ({key} = {value})"


class ERROR_AUTHENTICATE_FAILURE(ERROR_INVALID_ARGUMENT):
    _status_code = "UNAUTHENTICATED"
    _message = "Authenticate failure. (message = {message})"


class ERROR_PERMISSION_DENIED(ERROR_INVALID_ARGUMENT):
    _status_code = "PERMISSION_DENIED"
    _message = "Permission denied."


class ERROR_UNKNOWN(ERROR_BASE):
    _status_code = "INTERNAL"
    _message = "{message}"


class ERROR_REQUEST_TIMEOUT(ERROR_UNKNOWN):
    _status_code = "DEADLINE_EXCEEDED"
    _message = "Request timeout!"


class ERROR_TRANSACTION_STATUS(ERROR_UNKNOWN):
    _message = "Transaction status is incorrect. (status = {status})"


class ERROR_UNSUPPORTED_API(ERROR_UNKNOWN):
    _message = "Call Unsupported API. (reason = {reason})"


class ERROR_CONFIGURATION(ERROR_UNKNOWN):
    _message = "Configuration is invalid. (key = {key})"


class ERROR_DB_ENGINE_UNDEFINE(ERROR_CONFIGURATION):
    _message = "Database engine is undefined. (alias = {alias})"


class ERROR_CACHE_CONFIGURATION(ERROR_CONFIGURATION):
    _message = "Cache is not configured. (alias = {alias})"


class ERROR_CACHE_ENGINE_UNDEFINE(ERROR_CONFIGURATION):
    _message = "Cache engine is undefined. (alias = {alias})"


class ERROR_CONNECTOR_CONFIGURATION(ERROR_CONFIGURATION):
    _message = "Connector configuration is invalid. (backend = {backend})"


class ERROR_GRPC_CONFIGURATION(ERROR_CONFIGURATION):
    _message = "gRPC client configuration is invalid. ({endpoint}/{service}/{method})"


class ERROR_HANDLER_CONFIGURATION(ERROR_CONFIGURATION):
    _message = (
        "Handler configuration is invalid. (handler = {handler}, reason = {reason})"
    )


class ERROR_CONNECTOR_CONFIGURATION(ERROR_CONFIGURATION):
    _message = "Connector configuration is invalid. (connector = {connector}, reason = {reason})"


class ERROR_LOG_CONFIG(ERROR_CONFIGURATION):
    _message = "Log configuration is invalid. (reason = {reason})"


class ERROR_WRONG_CONFIGURATION(ERROR_CONFIGURATION):
    _message = "Configuration is invalid. ({key})"


class ERROR_DB_QUERY(ERROR_UNKNOWN):
    _message = "Database query failed. (reason = {reason})"


class ERROR_CACHE_OPTION(ERROR_UNKNOWN):
    _message = (
        "Does not support the cache option. (method = {method}, option = {option})"
    )


class ERROR_CACHE_TIMEOUT(ERROR_UNKNOWN):
    _message = "Cache timeout error. (config = {config})"


class ERROR_CACHE_ENCODE(ERROR_UNKNOWN):
    _message = "Cache data encoding failed. (reason = {reason})"


class ERROR_CACHE_DECODE(ERROR_UNKNOWN):
    _message = "Cache data decoding failed. (reason = {reason})"


class ERROR_CACHEABLE_VALUE_TYPE(ERROR_UNKNOWN):
    _message = "The value of cache.cacheable must be a dict type."


class ERROR_INTERNAL_API(ERROR_UNKNOWN):
    _message = "{message}"


class ERROR_UNAVAILAVBLE(ERROR_UNKNOWN):
    _status_code = "UNAVAILABLE"
    _message = "Server is unavailable. (reason = {reason})"


class ERROR_QUEUE_PUT(ERROR_UNAVAILAVBLE):
    _message = "Queue data put failed. (reason = {reason})"


class ERROR_QUEUE_GET(ERROR_UNAVAILAVBLE):
    _message = "Queue data get failed. (reason = {reason})"


class ERROR_GRPC_CONNECTION(ERROR_UNAVAILAVBLE):
    _status_code = "UNAVAILABLE"
    _message = "Server is unavailable. (channel = {channel}, message = {message})"


class ERROR_GRPC_TLS_HANDSHAKE(ERROR_GRPC_CONNECTION):
    _message = "TLS handshake failed. (reason = {reason})"


class ERROR_HANDLER(ERROR_UNKNOWN):
    _message = "'{handler_type} handler' import failed. (reason = {reason})"


class ERROR_CONNECTOR(ERROR_UNKNOWN):
    _message = "'{connector} handler' import failed. (reason = {reason})"


class ERROR_CONNECTOR_LOAD(ERROR_CONNECTOR):
    _message = "Failed to load connector. (connector = {connector}, reason = {reason})"


class ERROR_LOCATOR(ERROR_UNKNOWN):
    _message = "'{name}' load failed. (reason = {reason})"


class ERROR_TASK_LOCATOR(ERROR_LOCATOR):
    _message = "Call locator failure. (locator = {locator}, name = {name})"


class ERROR_TASK_METHOD(ERROR_UNKNOWN):
    _message = "Call method failure. (name = {name}, method = {method})"


class ERROR_NOT_IMPLEMENTED(ERROR_UNKNOWN):
    _status_code = "UNIMPLEMENTED"
    _message = "Not implemented."


class ERROR_MESSAGE_FORMAT(ERROR_UNKNOWN):
    _message = (
        "Error message format is invalid. (error_class = {error_class}, key = {key})"
    )


class ERROR_CACHE_KEY_FORMAT(ERROR_UNKNOWN):
    _message = "Cache key format is invalid. (key = {key})"


class ERROR_SERVICE_ACCOUNT_CANNOT_BE_DELETED_WITH_EXISTING_APP(ERROR_INVALID_ARGUMENT):
    _status_code = "INVALID_OPERATION"
    _message = """Service Account cannot be deleted as long as an associated App exists. 
    Please delete the App before deleting the Service Account. (error_class = {error_class}, key = {key})"""



import logging
import copy
import json
from fnmatch import fnmatch


class MaskingFilter(logging.Filter):
    def __init__(self, rules):
        self.rules = rules

    def filter(self, record):
        tnx_method = getattr(record, "tnx_method", None)
        parameter = getattr(record, "parameter", None)

        if tnx_method and parameter:
            record.parameter = self._check_masking(tnx_method, parameter)
        return True

    def _check_masking(self, tnx_method, params):
        masking_parameter = copy.deepcopy(params)

        if tnx_method in self.rules:
            return self._masking(masking_parameter, self.rules[tnx_method])

        """
        masking_parameter = copy.deepcopy(params)
        
        for _rule in self.rules:
            if fnmatch(tnx_method, _rule):
                masking_parameter = self._masking(masking_parameter, self.rules[_rule])
        """
        return masking_parameter

    @staticmethod
    def _masking(parameter, patterns):
        for _p in parameter:
            if _p in patterns:
                parameter[_p] = "********"

        return parameter



import re
import time
import random
import string
import secrets
import datetime
import yaml
import json
import hashlib
import urllib

from urllib.parse import urlparse
from dateutil.parser import isoparse
from typing import Tuple
from pathlib import Path
from typing import Union

YAML_LOADER = yaml.Loader
YAML_LOADER.add_implicit_resolver(
    "tag:yaml.org,2002:float",
    re.compile(
        """^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$""",
        re.X,
    ),
    list("-+0123456789."),
)


def generate_id(prefix: str = "id", nbytes: int = 6) -> str:
    random_id = secrets.token_hex(nbytes)
    return f"{prefix}-{random_id}"


def generate_trace_id() -> int:
    return random.getrandbits(128)


def generate_secret(nbytes: int = 32) -> str:
    return secrets.token_urlsafe(nbytes)


def generate_password(length: int = 8) -> str:
    # create alphanumerical from string constants
    printable = f"{string.ascii_letters}{string.digits}"

    # convert printable from string to list and shuffle
    printable = list(printable)
    random.shuffle(printable)

    # generate random password and convert to string
    random_password = random.choices(printable, k=length)

    # Append one lowercase and one uppercase and one number characters
    random_password.append(random.choices(list(string.ascii_lowercase))[0])
    random_password.append(random.choices(list(string.ascii_uppercase))[0])
    random_password.append(random.choices(list(string.digits))[0])

    random_password = "".join(random_password)
    return random_password


def random_string(nbytes: int = 6) -> str:
    return secrets.token_hex(nbytes)


def create_dir(path: str):
    p = Path(path)
    p.mkdir(parents=True, exist_ok=True)


def dump_json(data, indent=None, sort_keys=False) -> str:
    try:
        return json.dumps(data, indent=indent, sort_keys=sort_keys, ensure_ascii=False)
    except Exception as e:
        raise ValueError(f"JSON Dump Error: {str(e)}")


def save_json_to_file(data, json_file: str, indent=None, sort_keys=False):
    try:
        json_str = dump_json(data, indent, sort_keys)
        with open(json_file, "w") as f:
            f.write(json_str)
            f.close()
    except Exception as e:
        raise ValueError(f"JSON Save Error: {str(e)}")


def load_json(json_str: str) -> dict:
    try:
        return json.loads(json_str)
    except Exception:
        raise ValueError(f"JSON Load Error: {json_str}")


def load_json_from_file(json_file: str) -> dict:
    try:
        with open(json_file, "r") as f:
            return load_json(f.read())
    except Exception:
        raise Exception(f"JSON Load Error: {json_file}")


def dump_yaml(data: dict) -> str:
    try:
        return yaml.dump(data, allow_unicode=True)
    except Exception as e:
        raise ValueError(f"YAML Dump Error: {str(e)}")


def save_yaml_to_file(data: dict, yaml_file: str):
    try:
        yaml_str = dump_yaml(data)
        with open(yaml_file, "w") as f:
            f.write(yaml_str)
            f.close()
    except Exception as e:
        raise ValueError(f"YAML Save Error: {str(e)}")


def load_yaml(yaml_str: str) -> dict:
    try:
        return yaml.load(yaml_str, Loader=YAML_LOADER)
    except Exception:
        raise ValueError(f"YAML Load Error: {yaml_str}")


def load_yaml_from_file(yaml_file: str) -> dict:
    try:
        with open(yaml_file, "r") as f:
            return load_yaml(f.read())
    except Exception:
        raise Exception(f"YAML Load Error: {yaml_file}")


def load_yaml_from_url(url: str) -> dict:
    try:
        f = urllib.urlopen(url)
        return load_yaml(f.read())
    except Exception as e:
        raise Exception(f"Http call error: {url}. e={e}")


def parse_endpoint(endpoint: str) -> dict:
    try:
        o = urlparse(endpoint)
    except Exception:
        raise ValueError(f"Endpoint is invalid. ({endpoint})")

    return {"scheme": o.scheme, "hostname": o.hostname, "port": o.port, "path": o.path}


def parse_grpc_endpoint(endpoint: str) -> dict:
    try:
        endpoint_info = parse_endpoint(endpoint)

        if endpoint_info["scheme"] not in ["grpc", "grpc+ssl"]:
            raise ValueError(
                f"Unsupported protocol. (supported_protocol=grpc|grpc+ssl, endpoint={endpoint})"
            )

    except Exception:
        raise ValueError(f"gRPC Endpoint is invalid. (endpoint={endpoint})")

    if endpoint_info["scheme"] == "grpc+ssl":
        ssl_enabled = True
    else:
        ssl_enabled = False

    return {
        "endpoint": f'{endpoint_info["hostname"]}:{endpoint_info["port"]}',
        "ssl_enabled": ssl_enabled,
    }


def parse_grpc_uri(uri: str) -> dict:
    try:
        endpoint_info = parse_endpoint(uri)

        if endpoint_info["scheme"] not in ["grpc", "grpc+ssl"]:
            raise ValueError(
                f"Unsupported protocol. (supported_protocol=grpc|grpc+ssl, uri={uri})"
            )

        version, service, method = filter(
            lambda x: x.strip() != "", endpoint_info["path"].split("/")
        )
    except Exception:
        raise ValueError(f"gRPC URI is invalid. (uri={uri})")

    if endpoint_info["scheme"] == "grpc+ssl":
        ssl_enabled = True
    else:
        ssl_enabled = False

    return {
        "endpoint": f'{endpoint_info["hostname"]}:{endpoint_info["port"]}',
        "version": version,
        "service": service,
        "method": method,
        "ssl_enabled": ssl_enabled,
    }


def deep_merge(from_dict: dict, into_dict: dict) -> dict:
    for key, value in from_dict.items():
        if isinstance(value, dict):
            node = into_dict.setdefault(key, {})
            deep_merge(value, node)
        else:
            into_dict[key] = value

    return into_dict


def parse_timediff_query(query: str) -> datetime:
    try:
        time_info, include_operator = _parse_timediff_from_regex(query)
        base_dt = _convert_base_time(time_info["base_time"])
        if include_operator:
            time_delta = _convert_time_delta(
                time_info["time_delta_number"], time_info["time_delta_unit"]
            )
            if time_info["operator"] == "+":
                return base_dt + time_delta
            else:
                return base_dt - time_delta
        else:
            return base_dt
    except Exception as e:
        raise ValueError(f"Timediff format is invalid. (value={query})")


def _convert_time_delta(
    time_delta_number: str, time_delta_unit: str
) -> datetime.timedelta:
    _time_delta_map = {
        "s": "seconds",
        "m": "minutes",
        "h": "hours",
        "d": "days",
        "w": "weeks",
    }

    time_delta_params = {_time_delta_map[time_delta_unit]: int(time_delta_number)}

    return datetime.timedelta(**time_delta_params)


def _convert_base_time(time_str: str) -> datetime:
    now = datetime.datetime.utcnow()
    if time_str == "now":
        return now
    elif time_str == "now/d":
        return datetime.datetime.combine(now.date(), datetime.time(0))
    elif time_str == "now/w":
        today = datetime.datetime.combine(now.date(), datetime.time(0))
        return today - datetime.timedelta(days=now.date().weekday())
    elif time_str == "now/m":
        today = datetime.datetime.combine(now.date(), datetime.time(0))
        return today.replace(day=1)


def _parse_timediff_from_regex(query: str) -> Tuple[dict, bool]:
    p = (
        r"^\s?(?P<base_time>now(\/[dwm])?)\s?"
        r"(?P<operator>[+|-])\s?"
        r"(?P<time_delta_number>\d+)"
        r"(?P<time_delta_unit>[s|m|h|d|w])\s?$"
    )
    rule = re.compile(p)
    match = rule.match(query)
    if match:
        return match.groupdict(), True
    else:
        if query.strip() not in ["now", "now/d", "now/w", "now/m"]:
            raise ValueError(f"Timediff format is invalid. (value={query})")

        return {"base_time": query.strip()}, False


def get_dict_value(data: dict, dotted_key: str, default_value: any = None):
    if "." in dotted_key:
        key, rest = dotted_key.split(".", 1)

        if isinstance(data, dict) and key in data:
            if isinstance(data[key], list):
                list_values = []
                for value in data[key]:
                    list_values.append(get_dict_value(value, rest, default_value))

                return list_values

            else:
                return get_dict_value(data[key], rest, default_value)

        else:
            return default_value
    else:
        if isinstance(data, dict):
            return data.get(dotted_key, default_value)
        else:
            return default_value


def get_list_values(values: list, dotted_key: str, default_value: any = None):
    match_option = "contain"
    if len(dotted_key) > 1 and dotted_key[0] == "?":
        condition = True
        try:
            cond_option, rest = dotted_key[1:].split("=>", 1)
            cond_key, cond_value = cond_option.split(":", 1)
        except Exception as e:
            # Syntax Error
            return default_value

        if cond_value[:1] == "=":
            match_option = "eq"
            cond_value = cond_value[1:]

        elif cond_value[:1] == "!":
            match_option = "not"
            cond_value = cond_value[1:]

    else:
        try:
            # Get value by index
            if "." in dotted_key:
                index, rest = dotted_key.split(".", 1)
                index = int(index)
                if index >= len(values):
                    return default_value
                else:
                    if rest.strip() != "" and isinstance(values[index], dict):
                        return get_dict_value(values[index], rest, default_value)
                    else:
                        values = values[index]
            else:
                return values[int(dotted_key)]

        except Exception:
            rest = dotted_key

        condition = False
        cond_key = None
        cond_value = None

    results = []
    for value in values:
        # Get value from condition
        if not isinstance(value, dict) or (
            condition
            and not _check_condition(match_option, value[cond_key], cond_value)
        ):
            continue

        # Get value from dict key
        result = get_dict_value(value, rest, default_value)

        if result:
            if isinstance(result, list):
                results += result
            else:
                results.append(result)

    try:
        if len(results) > 0:
            return list(set(results))
        else:
            return default_value
    except Exception:
        return results


def _check_condition(match_option: str, val1, val2):
    val1 = str(val1).lower()
    val2 = str(val2).lower()

    if match_option == "eq":
        if val1 == val2:
            return True

    elif match_option == "not":
        if val1.find(val2) < 0:
            return True

    else:
        if val1.find(val2) >= 0:
            return True

    return False


def change_dict_value(
    data: dict,
    dotted_key: str,
    change_value,
    change_type="value",
    allow_new_key: bool = False,
) -> dict:
    # change_value = func or value(any type)
    if "." in dotted_key:
        key, rest = dotted_key.split(".", 1)
        if key in data:
            if rest.startswith("[]") and isinstance(data[key], list):
                list_data = []
                for sub_data in data[key]:
                    if rest.strip() == "[]":
                        list_data.append(
                            _change_value_by_type(change_type, sub_data, change_value)
                        )
                    else:
                        sub_rest = rest.split(".", 1)[1]
                        list_data.append(
                            change_dict_value(
                                sub_data,
                                sub_rest,
                                change_value,
                                change_type,
                                allow_new_key=allow_new_key,
                            )
                        )
                data[key] = list_data
            elif isinstance(data[key], dict):
                data[key] = change_dict_value(
                    data[key],
                    rest,
                    change_value,
                    change_type,
                    allow_new_key=allow_new_key,
                )
    else:
        if dotted_key in data:
            data[dotted_key] = _change_value_by_type(
                change_type, data[dotted_key], change_value
            )
        else:
            if allow_new_key:
                data[dotted_key] = change_value

    return data


def _change_value_by_type(change_type, original_value, change_value):
    if change_type == "value":
        return change_value
    elif change_type == "func":
        return change_value(original_value)
    else:
        return original_value


def date_to_string(
    value: datetime.date, date_format: str = "%Y-%m-%d"
) -> Union[str, None]:
    if isinstance(value, datetime.date):
        return value.strftime(date_format)

    return None


def datetime_to_iso8601(value: datetime.datetime) -> Union[str, None]:
    if isinstance(value, datetime.datetime):
        value = value.replace(tzinfo=None)
        return f"{value.isoformat(timespec='milliseconds')}Z"

    return None


def iso8601_to_datetime(value: str) -> Union[datetime.datetime, None]:
    if isinstance(value, str):
        try:
            return isoparse(value)
        except Exception as e:
            raise ValueError(f"Datetime(ISO8601) format is invalid. (value={value})")

    return None


def iso8601_to_timestamp(value: str) -> Union[int, None]:
    dt = iso8601_to_datetime(value)
    if dt:
        return int(time.mktime(dt.timetuple()))

    return None


def tags_to_dict(tags: list) -> Union[dict, None]:
    if isinstance(tags, list):
        dict_value = {}
        for tag in tags:
            dict_value[tag["key"]] = tag["value"]

        return dict_value

    else:
        return None


def dict_to_tags(dict_value: dict) -> list:
    tags = []
    if isinstance(dict_value, dict):
        for key, value in dict_value.items():
            tags.append({"key": key, "value": value})

    return tags


def dict_to_hash(dict_value: dict) -> str:
    """MD5 hash of a dictionary."""
    dhash = hashlib.md5()
    encoded = json.dumps(dict_value, sort_keys=True).encode()
    dhash.update(encoded)
    return dhash.hexdigest()


def string_to_hash(str_value: str) -> str:
    """MD5 hash of a String."""
    dhash = hashlib.md5(str_value.encode("utf-8"))
    return dhash.hexdigest()


def change_dict_with_dot_notation(dict_value: dict, key="", dots=None) -> dict:
    if dots is None:
        dots = {}
    if isinstance(dict_value, dict):
        for k, v in dict_value.items():
            change_dict_with_dot_notation(
                dict_value[k], f"{key}.{k}" if key else k, dots
            )
    else:
        dots[key] = dict_value
    return dots


if __name__ == "__main__":
    pass



from spaceone.core.locator import Locator
from spaceone.core.transaction import Transaction, get_transaction


class CoreObject(object):

    def __init__(self, *args, **kwargs):
        self.locator = Locator()
        get_transaction()

    @property
    def transaction(self) -> Transaction:
        return get_transaction()



import logging


class TracebackFilter(logging.Filter):
    def filter(self, record):
        _traceback = getattr(record, "traceback", None)

        if _traceback is None or _traceback == "":
            record.traceback = ""
        else:
            record.traceback = (
                _traceback.replace("\\n", "\n").replace('\\"', '"').replace("\\t", "\t")
            )

        return True


class TracebackLogFilter(logging.Filter):
    def filter(self, record):
        _traceback = getattr(record, "traceback", None)

        if _traceback is None or _traceback == "":
            record.traceback_log = '""'
        else:
            record.traceback_log = _traceback

        return True
